""" Base class to implement a new LLM

This module is the base class to integrate the various LLMs API. This module also
includes the Base LLM classes for OpenAI and Google PaLM.

Example:

    ```
    from .base import BaseOpenAI

    class CustomLLM(BaseOpenAI):

        Custom Class Starts here!!
    ```
"""

from __future__ import annotations

import ast
import re
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Tuple, Union

from pandasai.helpers.memory import Memory
from pandasai.prompts.generate_system_message import GenerateSystemMessagePrompt

from ..exceptions import (
    APIKeyNotFoundError,
    MethodNotImplementedError,
    NoCodeFoundError,
)
from ..helpers.openai import is_openai_v1
from ..helpers.openai_info import openai_callback_var
from ..prompts.base import BasePrompt

if TYPE_CHECKING:
    from pandasai.pipelines.pipeline_context import PipelineContext


class LLM:
    """Base class to implement a new LLM."""

    last_prompt: Optional[str] = None

    def is_pandasai_llm(self) -> bool:
        """
        Return True if the LLM is from pandasAI.

        Returns:
            bool: True if the LLM is from pandasAI

        """
        return True

    @property
    def type(self) -> str:
        """
        Return type of LLM.

        Raises:
            APIKeyNotFoundError: Type has not been implemented

        Returns:
            str: Type of LLM a string

        """
        raise APIKeyNotFoundError("Type has not been implemented")

    def _polish_code(self, code: str) -> str:
        """
        Polish the code by removing the leading "python" or "py",  \
        removing surrounding '`' characters  and removing trailing spaces and new lines.

        Args:
            code (str): A string of Python code.

        Returns:
            str: Polished code.

        """
        if re.match(r"^(python|py)", code):
            code = re.sub(r"^(python|py)", "", code)
        if re.match(r"^`.*`$", code):
            code = re.sub(r"^`(.*)`$", r"\1", code)
        code = code.strip()
        return code

    def _is_python_code(self, string):
        """
        Return True if it is valid python code.
        Args:
            string (str):

        Returns (bool): True if Python Code otherwise False

        """
        try:
            ast.parse(string)
            return True
        except SyntaxError:
            return False

    def _extract_code(self, response: str, separator: str = "```") -> str:
        """
        Extract the code from the response.

        Args:
            response (str): Response
            separator (str, optional): Separator. Defaults to "```".

        Raises:
            NoCodeFoundError: No code found in the response

        Returns:
            str: Extracted code from the response

        """
        code = response

        # If separator is in the response then we want the code in between only
        if separator in response and len(code.split(separator)) > 1:
            code = code.split(separator)[1]
        code = self._polish_code(code)

        # Even if the separator is not in the response, the output might still be valid python code
        if not self._is_python_code(code):
            raise NoCodeFoundError("No code found in the response")

        return code

    def prepend_system_prompt(self, prompt: BasePrompt, memory: Memory):
        """
        Append system prompt to the chat prompt, useful when model doesn't have messages for chat history
        Args:
            prompt (BasePrompt): prompt for chat method
            memory (Memory): user conversation history
        """
        return self.get_system_prompt(memory) + prompt if memory else prompt

    def get_system_prompt(self, memory: Memory) -> Any:
        """
        Generate system prompt with agent info and previous conversations
        """
        system_prompt = GenerateSystemMessagePrompt(memory=memory)
        return system_prompt.to_string()

    def get_messages(self, memory: Memory) -> Any:
        """
        Return formatted messages
        Args:
            memory (Memory): Get past Conversation from memory
        """
        return memory.get_previous_conversation()

    def _extract_tag_text(self, response: str, tag: str) -> str:
        """
        Extracts the text between two tags in the response.

        Args:
            response (str): Response
            tag (str): Tag name

        Returns:
            (str or None): Extracted text from the response
        """

        if match := re.search(
            f"(<{tag}>)(.*)(</{tag}>)",
            response,
            re.DOTALL | re.MULTILINE,
        ):
            return match[2]
        return None

    @abstractmethod
    def call(self, instruction: BasePrompt, context: PipelineContext = None) -> str:
        """
        Execute the LLM with given prompt.

        Args:
            instruction (BasePrompt): A prompt object with instruction for LLM.
            context (PipelineContext, optional): PipelineContext. Defaults to None.

        Raises:
            MethodNotImplementedError: Call method has not been implemented

        """
        raise MethodNotImplementedError("Call method has not been implemented")

    def generate_code(self, instruction: BasePrompt, context: PipelineContext) -> str:
        """
        Generate the code based on the instruction and the given prompt.

        Args:
            instruction (BasePrompt): Prompt with instruction for LLM.

        Returns:
            str: A string of Python code.

        """
        response = self.call(instruction, context)
        return self._extract_code(response)


class BaseOpenAI(LLM):
    """Base class to implement a new OpenAI LLM.

    LLM base class, this class is extended to be used with OpenAI API.

    """

    api_token: str
    api_base: str = "https://api.openai.com/v1"
    temperature: float = 0
    max_tokens: int = 1000
    top_p: float = 1
    frequency_penalty: float = 0
    presence_penalty: float = 0.6
    best_of: int = 1
    n: int = 1
    stop: Optional[str] = None
    request_timeout: Union[float, Tuple[float, float], Any, None] = None
    max_retries: int = 2
    seed: Optional[int] = None
    # support explicit proxy for OpenAI
    openai_proxy: Optional[str] = None
    default_headers: Union[Mapping[str, str], None] = None
    default_query: Union[Mapping[str, object], None] = None
    # Configure a custom httpx client. See the
    # [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
    http_client: Union[Any, None] = None
    client: Any
    _is_chat_model: bool

    def _set_params(self, **kwargs):
        """
        Set Parameters
        Args:
            **kwargs: ["model", "deployment_name", "temperature","max_tokens",
            "top_p", "frequency_penalty", "presence_penalty", "stop", "seed"]

        Returns:
            None.

        """

        valid_params = [
            "model",
            "deployment_name",
            "temperature",
            "max_tokens",
            "top_p",
            "frequency_penalty",
            "presence_penalty",
            "stop",
            "seed",
        ]
        for key, value in kwargs.items():
            if key in valid_params:
                setattr(self, key, value)

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling OpenAI API."""
        params: Dict[str, Any] = {
            "temperature": self.temperature,
            "top_p": self.top_p,
            "frequency_penalty": self.frequency_penalty,
            "presence_penalty": self.presence_penalty,
            "seed": self.seed,
            "stop": self.stop,
            "n": self.n,
        }

        if self.max_tokens is not None:
            params["max_tokens"] = self.max_tokens

        # Azure gpt-35-turbo doesn't support best_of
        # don't specify best_of if it is 1
        if self.best_of > 1:
            params["best_of"] = self.best_of

        return params

    @property
    def _invocation_params(self) -> Dict[str, Any]:
        """Get the parameters used to invoke the model."""
        openai_creds: Dict[str, Any] = {}
        if not is_openai_v1():
            openai_creds |= {
                "api_key": self.api_token,
                "api_base": self.api_base,
            }

        return {**openai_creds, **self._default_params}

    @property
    def _client_params(self) -> Dict[str, any]:
        return {
            "api_key": self.api_token,
            "base_url": self.api_base,
            "timeout": self.request_timeout,
            "max_retries": self.max_retries,
            "default_headers": self.default_headers,
            "default_query": self.default_query,
            "http_client": self.http_client,
        }

    def completion(self, prompt: str, memory: Memory) -> str:
        """
        Query the completion API

        Args:
            prompt (str): A string representation of the prompt.

        Returns:
            str: LLM response.

        """
        prompt = self.prepend_system_prompt(prompt, memory)

        params = {**self._invocation_params, "prompt": prompt}

        if self.stop is not None:
            params["stop"] = [self.stop]

        response = self.client.create(**params)

        if openai_handler := openai_callback_var.get():
            openai_handler(response)

        self.last_prompt = prompt

        return response.choices[0].text

    def chat_completion(self, value: str, memory: Memory) -> str:
        """
        Query the chat completion API

        Args:
            value (str): Prompt

        Returns:
            str: LLM response.

        """
        messages = memory.to_openai_messages() if memory else []

        # adding current prompt as latest query message
        messages.append(
            {
                "role": "user",
                "content": value,
            },
        )

        params = {
            **self._invocation_params,
            "messages": messages,
        }

        if self.stop is not None:
            params["stop"] = [self.stop]

        response = self.client.create(**params)

        if openai_handler := openai_callback_var.get():
            openai_handler(response)

        return response.choices[0].message.content

    def call(self, instruction: BasePrompt, context: PipelineContext = None):
        """
        Call the OpenAI LLM.

        Args:
            instruction (BasePrompt): A prompt object with instruction for LLM.
            context (PipelineContext): context to pass.

        Raises:
            UnsupportedModelError: Unsupported model

        Returns:
            str: Response
        """
        self.last_prompt = instruction.to_string()

        memory = context.memory if context else None

        return (
            self.chat_completion(self.last_prompt, memory)
            if self._is_chat_model
            else self.completion(self.last_prompt, memory)
        )


class BaseGoogle(LLM):
    """Base class to implement a new Google LLM

    LLM base class is extended to be used with
    """

    temperature: Optional[float] = 0
    top_p: Optional[float] = 0.8
    top_k: Optional[int] = 40
    max_output_tokens: Optional[int] = 1000

    def _valid_params(self):
        return ["temperature", "top_p", "top_k", "max_output_tokens"]

    def _set_params(self, **kwargs):
        """
        Dynamically set Parameters for the object.

        Args:
            **kwargs:
                Possible keyword arguments: "temperature", "top_p", "top_k",
                "max_output_tokens".

        Returns:
            None.

        """

        valid_params = self._valid_params()
        for key, value in kwargs.items():
            if key in valid_params:
                setattr(self, key, value)

    def _validate(self):
        """Validates the parameters for Google"""

        if self.temperature is not None and not 0 <= self.temperature <= 1:
            raise ValueError("temperature must be in the range [0.0, 1.0]")

        if self.top_p is not None and not 0 <= self.top_p <= 1:
            raise ValueError("top_p must be in the range [0.0, 1.0]")

        if self.top_k is not None and not 0 <= self.top_k <= 100:
            raise ValueError("top_k must be in the range [0.0, 100.0]")

        if self.max_output_tokens is not None and self.max_output_tokens <= 0:
            raise ValueError("max_output_tokens must be greater than zero")

    @abstractmethod
    def _generate_text(self, prompt: str, memory: Optional[Memory] = None) -> str:
        """
        Generates text for prompt, specific to implementation.

        Args:
            prompt (str): A string representation of the prompt.

        Returns:
            str: LLM response.

        """
        raise MethodNotImplementedError("method has not been implemented")

    def call(self, instruction: BasePrompt, context: PipelineContext = None) -> str:
        """
        Call the Google LLM.

        Args:
            instruction (BasePrompt): Instruction to pass.
            context (PipelineContext): Pass PipelineContext.

        Returns:
            str: LLM response.

        """
        self.last_prompt = instruction.to_string()
        memory = context.memory if context else None
        return self._generate_text(self.last_prompt, memory)



from typing import Optional

from pandasai.agent.base_judge import BaseJudge
from pandasai.ee.agents.semantic_agent.pipeline.code_generator import CodeGenerator
from pandasai.ee.agents.semantic_agent.pipeline.error_correction_pipeline.error_correction_pipeline import (
    ErrorCorrectionPipeline,
)
from pandasai.ee.agents.semantic_agent.pipeline.error_correction_pipeline.fix_semantic_json_pipeline import (
    FixSemanticJsonPipeline,
)
from pandasai.ee.agents.semantic_agent.pipeline.llm_call import LLMCall
from pandasai.ee.agents.semantic_agent.pipeline.Semantic_prompt_generation import (
    SemanticPromptGeneration,
)
from pandasai.ee.agents.semantic_agent.pipeline.semantic_result_parsing import (
    SemanticResultParser,
)
from pandasai.ee.agents.semantic_agent.pipeline.validate_pipeline_input import (
    ValidatePipelineInput,
)
from pandasai.helpers.logger import Logger
from pandasai.pipelines.chat.cache_lookup import CacheLookup
from pandasai.pipelines.chat.code_cleaning import CodeCleaning
from pandasai.pipelines.chat.code_execution import CodeExecution
from pandasai.pipelines.chat.error_correction_pipeline.error_correction_pipeline_input import (
    ErrorCorrectionPipelineInput,
)
from pandasai.pipelines.chat.generate_chat_pipeline import GenerateChatPipeline
from pandasai.pipelines.chat.result_validation import ResultValidation
from pandasai.pipelines.pipeline import Pipeline
from pandasai.pipelines.pipeline_context import PipelineContext


class SemanticChatPipeline(GenerateChatPipeline):
    code_generation_pipeline = Pipeline
    code_execution_pipeline = Pipeline
    context: PipelineContext
    _logger: Logger
    last_error: str

    def __init__(
        self,
        context: Optional[PipelineContext] = None,
        logger: Optional[Logger] = None,
        judge: BaseJudge = None,
        on_prompt_generation=None,
        on_code_generation=None,
        before_code_execution=None,
        on_result=None,
    ):
        super().__init__(
            context,
            logger,
            judge=judge,
            on_prompt_generation=on_prompt_generation,
            on_code_generation=on_code_generation,
            before_code_execution=before_code_execution,
            on_result=on_result,
        )

        self.code_generation_pipeline = Pipeline(
            context=context,
            logger=logger,
            query_exec_tracker=self.query_exec_tracker,
            steps=[
                ValidatePipelineInput(),
                CacheLookup(),
                SemanticPromptGeneration(
                    skip_if=self.is_cached,
                    on_execution=on_prompt_generation,
                ),
                LLMCall(),
                CodeGenerator(
                    on_execution=on_code_generation,
                    on_failure=self.on_wrong_semantic_json,
                ),
                CodeCleaning(
                    skip_if=self.no_code,
                    on_failure=self.on_code_cleaning_failure,
                    on_retry=self.on_code_retry,
                ),
            ],
        )

        self.code_execution_pipeline = Pipeline(
            context=context,
            logger=logger,
            query_exec_tracker=self.query_exec_tracker,
            steps=[
                CodeExecution(
                    before_execution=before_code_execution,
                    on_failure=self.on_code_execution_failure,
                    on_retry=self.on_code_retry,
                ),
                ResultValidation(),
                SemanticResultParser(
                    before_execution=on_result,
                ),
            ],
        )

        self.code_exec_error_pipeline = ErrorCorrectionPipeline(
            context=context,
            logger=logger,
            query_exec_tracker=self.query_exec_tracker,
            on_code_generation=on_code_generation,
            on_prompt_generation=on_prompt_generation,
        )

        self.fix_semantic_json_pipeline = FixSemanticJsonPipeline(
            context=context,
            logger=logger,
            query_exec_tracker=self.query_exec_tracker,
            on_code_generation=on_code_generation,
            on_prompt_generation=on_prompt_generation,
        )

        self.context = context
        self._logger = logger
        self.last_error = None

    def on_wrong_semantic_json(self, code, errors):
        self.query_exec_tracker.add_step(
            {
                "type": "CodeGenerator",
                "success": False,
                "message": "Failed to validate json",
                "execution_time": None,
                "data": {
                    "content_type": "code",
                    "value": code,
                    "exception": errors,
                },
            }
        )
        correction_input = ErrorCorrectionPipelineInput(code, errors)
        return self.fix_semantic_json_pipeline.run(correction_input)



import json
import os
import uuid
from typing import List, Optional, Union

import pandasai.pandas as pd
from pandasai.agent.base_security import BaseSecurity
from pandasai.llm.bamboo_llm import BambooLLM
from pandasai.pipelines.chat.chat_pipeline_input import ChatPipelineInput
from pandasai.pipelines.chat.code_execution_pipeline_input import (
    CodeExecutionPipelineInput,
)
from pandasai.vectorstores.vectorstore import VectorStore

from ..config import load_config_from_json
from ..connectors import BaseConnector, PandasConnector
from ..constants import DEFAULT_CACHE_DIRECTORY, DEFAULT_CHART_DIRECTORY
from ..exceptions import (
    InvalidLLMOutputType,
    MaliciousQueryError,
    MissingVectorStoreError,
)
from ..helpers.df_info import df_type
from ..helpers.folder import Folder
from ..helpers.logger import Logger
from ..helpers.memory import Memory
from ..llm.base import LLM
from ..llm.langchain import LangchainLLM, is_langchain_llm
from ..pipelines.pipeline_context import PipelineContext
from ..prompts.base import BasePrompt
from ..prompts.clarification_questions_prompt import ClarificationQuestionPrompt
from ..prompts.explain_prompt import ExplainPrompt
from ..prompts.rephase_query_prompt import RephraseQueryPrompt
from ..schemas.df_config import Config
from ..skills import Skill
from .callbacks import Callbacks


class BaseAgent:
    """
    Base Agent class to improve the conversational experience in PandasAI
    """

    def __init__(
        self,
        dfs: Union[
            pd.DataFrame, BaseConnector, List[Union[pd.DataFrame, BaseConnector]]
        ],
        config: Optional[Union[Config, dict]] = None,
        memory_size: Optional[int] = 10,
        vectorstore: Optional[VectorStore] = None,
        description: str = None,
        security: BaseSecurity = None,
    ):
        """
        Args:
            df (Union[pd.DataFrame, List[pd.DataFrame]]): Pandas or Modin dataframe
            Polars or Database connectors
            memory_size (int, optional): Conversation history to use during chat.
            Defaults to 1.
        """
        self.last_prompt = None
        self.last_prompt_id = None
        self.last_result = None
        self.last_code_generated = None
        self.last_code_executed = None
        self.agent_info = description

        self.conversation_id = uuid.uuid4()

        self.dfs = self.get_dfs(dfs)

        # Instantiate the context
        self.config = self.get_config(config)
        self.context = PipelineContext(
            dfs=self.dfs,
            config=self.config,
            memory=Memory(memory_size, agent_info=description),
            vectorstore=vectorstore,
        )

        # Instantiate the logger
        self.logger = Logger(
            save_logs=self.config.save_logs, verbose=self.config.verbose
        )

        # Instantiate the vectorstore
        self._vectorstore = vectorstore

        if self._vectorstore is None and os.environ.get("PANDASAI_API_KEY"):
            try:
                from pandasai.vectorstores.bamboo_vectorstore import BambooVectorStore
            except ImportError as e:
                raise ImportError(
                    "Could not import BambooVectorStore. Please install the required dependencies."
                ) from e

            self._vectorstore = BambooVectorStore(logger=self.logger)
            self.context.vectorstore = self._vectorstore

        self._callbacks = Callbacks(self)

        self.configure()

        self.pipeline = None
        self.security = security

    def configure(self):
        # Add project root path if save_charts_path is default
        if (
            self.config.save_charts
            and self.config.save_charts_path == DEFAULT_CHART_DIRECTORY
        ):
            Folder.create(self.config.save_charts_path)

        # Add project root path if cache_path is default
        if self.config.enable_cache:
            Folder.create(DEFAULT_CACHE_DIRECTORY)

    def get_config(self, config: Union[Config, dict]):
        """
        Load a config to be used to run the queries.

        Args:
            config (Union[Config, dict]): Config to be used
        """

        config = load_config_from_json(config)

        if isinstance(config, dict) and config.get("llm") is not None:
            config["llm"] = self.get_llm(config["llm"])

        config = Config(**config)

        if config.llm is None:
            config.llm = BambooLLM()

        return config

    def get_llm(self, llm: LLM) -> LLM:
        """
        Load a LLM to be used to run the queries.
        Check if it is a PandasAI LLM or a Langchain LLM.
        If it is a Langchain LLM, wrap it in a PandasAI LLM.

        Args:
            llm (object): LLMs option to be used for API access

        Raises:
            BadImportError: If the LLM is a Langchain LLM but the langchain package
            is not installed
        """
        if is_langchain_llm(llm):
            llm = LangchainLLM(llm)

        return llm

    def get_dfs(
        self,
        dfs: Union[
            pd.DataFrame, BaseConnector, List[Union[pd.DataFrame, BaseConnector]]
        ],
    ):
        """
        Load all the dataframes to be used in the agent.

        Args:
            dfs (List[Union[pd.DataFrame, Any]]): Pandas dataframe
        """
        # Inline import to avoid circular import
        from pandasai.smart_dataframe import SmartDataframe

        # If only one dataframe is passed, convert it to a list
        if not isinstance(dfs, list):
            dfs = [dfs]

        connectors = []
        for df in dfs:
            if isinstance(df, BaseConnector):
                connectors.append(df)
            elif isinstance(df, (pd.DataFrame, pd.Series, list, dict, str)):
                connectors.append(PandasConnector({"original_df": df}))
            elif df_type(df) == "modin":
                connectors.append(PandasConnector({"original_df": df}))
            elif isinstance(df, SmartDataframe) and isinstance(
                df.dataframe, BaseConnector
            ):
                connectors.append(df.dataframe)
            else:
                try:
                    import polars as pl

                    if isinstance(df, pl.DataFrame):
                        from ..connectors.polars import PolarsConnector

                        connectors.append(PolarsConnector({"original_df": df}))

                    else:
                        raise ValueError(
                            "Invalid input data. We cannot convert it to a dataframe."
                        )
                except ImportError as e:
                    raise ValueError(
                        "Invalid input data. We cannot convert it to a dataframe."
                    ) from e
        return connectors

    def add_skills(self, *skills: Skill):
        """
        Add Skills to PandasAI
        """
        self.context.skills_manager.add_skills(*skills)

    def call_llm_with_prompt(self, prompt: BasePrompt):
        """
        Call LLM with prompt using error handling to retry based on config
        Args:
            prompt (BasePrompt): BasePrompt to pass to LLM's
        """
        retry_count = 0
        while retry_count < self.context.config.max_retries:
            try:
                result: str = self.context.config.llm.call(prompt)
                if prompt.validate(result):
                    return result
                else:
                    raise InvalidLLMOutputType("Response validation failed!")
            except Exception:
                if (
                    not self.context.config.use_error_correction_framework
                    or retry_count >= self.context.config.max_retries - 1
                ):
                    raise
                retry_count += 1

    def check_malicious_keywords_in_query(self, query):
        dangerous_modules = [
            " os",
            " io",
            ".os",
            ".io",
            "'os'",
            "'io'",
            '"os"',
            '"io"',
            "chr(",
            "chr)",
            "chr ",
            "(chr",
            "b64decode",
        ]
        return any(module in query for module in dangerous_modules)

    def chat(self, query: str, output_type: Optional[str] = None):
        """
        Simulate a chat interaction with the assistant on Dataframe.
        """
        if not self.pipeline:
            return (
                "Unfortunately, I was not able to get your answers, "
                "because of the following error: No pipeline exists"
            )

        try:
            self.logger.log(f"Question: {query}")
            self.logger.log(
                f"Running PandasAI with {self.context.config.llm.type} LLM..."
            )

            self.assign_prompt_id()

            if self.check_malicious_keywords_in_query(query):
                raise MaliciousQueryError(
                    "The query contains references to io or os modules or b64decode method which can be used to execute or access system resources in unsafe ways."
                )

            if self.security and self.security.evaluate(query):
                raise MaliciousQueryError("Query can result in a malicious code")

            pipeline_input = ChatPipelineInput(
                query, output_type, self.conversation_id, self.last_prompt_id
            )

            return self.pipeline.run(pipeline_input)

        except Exception as exception:
            return (
                "Unfortunately, I was not able to get your answers, "
                "because of the following error:\n"
                f"\n{exception}\n"
            )

    def generate_code(self, query: str, output_type: Optional[str] = None):
        """
        Simulate code generation with the assistant on Dataframe.
        """
        if not self.pipeline:
            return (
                "Unfortunately, I was not able to get your answers, "
                "because of the following error: No pipeline exists"
            )
        try:
            self.logger.log(f"Question: {query}")
            self.logger.log(
                f"Running PandasAI with {self.context.config.llm.type} LLM..."
            )

            self.assign_prompt_id()

            pipeline_input = ChatPipelineInput(
                query, output_type, self.conversation_id, self.last_prompt_id
            )

            return self.pipeline.run_generate_code(pipeline_input)
        except Exception as exception:
            return (
                "Unfortunately, I was not able to get your answers, "
                "because of the following error:\n"
                f"\n{exception}\n"
            )

    def execute_code(
        self, code: Optional[str] = None, output_type: Optional[str] = None
    ):
        """
        Execute code Generated with the assistant on Dataframe.
        """
        if not self.pipeline:
            return (
                "Unfortunately, I was not able to get your answers, "
                "because of the following error: No pipeline exists to execute try Agent class"
            )
        try:
            if code is None:
                code = self.last_code_generated
            self.logger.log(f"Code: {code}")
            self.logger.log(
                f"Running PandasAI with {self.context.config.llm.type} LLM..."
            )

            self.assign_prompt_id()

            pipeline_input = CodeExecutionPipelineInput(
                code, output_type, self.conversation_id, self.last_prompt_id
            )

            return self.pipeline.run_execute_code(pipeline_input)
        except Exception as exception:
            return (
                "Unfortunately, I was not able to get your answers, "
                "because of the following error:\n"
                f"\n{exception}\n"
            )

    def train(
        self,
        queries: Optional[List[str]] = None,
        codes: Optional[List[str]] = None,
        docs: Optional[List[str]] = None,
    ) -> None:
        """
        Trains the context to be passed to model
        Args:
            queries (Optional[str], optional): user user
            codes (Optional[str], optional): generated code
            docs (Optional[List[str]], optional): additional docs
        Raises:
            ImportError: if default vector db lib is not installed it raises an error
        """
        if self._vectorstore is None:
            raise MissingVectorStoreError(
                "No vector store provided. Please provide a vector store to train the agent."
            )

        if (queries and not codes) or (not queries and codes):
            raise ValueError(
                "If either queries or codes are provided, both must be provided."
            )

        if docs is not None:
            self._vectorstore.add_docs(docs)

        if queries and codes:
            self._vectorstore.add_question_answer(queries, codes)

        self.logger.log("Agent successfully trained on the data")

    def clear_memory(self):
        """
        Clears the memory
        """
        self.context.memory.clear()
        self.conversation_id = uuid.uuid4()

    def add_message(self, message, is_user=False):
        """
        Add message to the memory. This is useful when you want to add a message
        to the memory without calling the chat function (for example, when you
        need to add a message from the agent).
        """
        self.context.memory.add(message, is_user=is_user)

    def assign_prompt_id(self):
        """Assign a prompt ID"""

        self.last_prompt_id = uuid.uuid4()

        if self.logger:
            self.logger.log(f"Prompt ID: {self.last_prompt_id}")

    def clarification_questions(self, query: str) -> List[str]:
        """
        Generate clarification questions based on the data
        """
        prompt = ClarificationQuestionPrompt(
            context=self.context,
            query=query,
        )

        result = self.call_llm_with_prompt(prompt)
        self.logger.log(
            f"""Clarification Questions:  {result}
            """
        )
        result = result.replace("```json", "").replace("```", "")
        questions: list[str] = json.loads(result)
        return questions[:3]

    def start_new_conversation(self):
        """
        Clears the previous conversation
        """
        self.clear_memory()

    def explain(self) -> str:
        """
        Returns the explanation of the code how it reached to the solution
        """
        try:
            prompt = ExplainPrompt(
                context=self.context,
                code=self.last_code_executed,
            )
            response = self.call_llm_with_prompt(prompt)
            self.logger.log(
                f"""Explanation:  {response}
                """
            )
            return response
        except Exception as exception:
            return (
                "Unfortunately, I was not able to explain, "
                "because of the following error:\n"
                f"\n{exception}\n"
            )

    def rephrase_query(self, query: str):
        try:
            prompt = RephraseQueryPrompt(
                context=self.context,
                query=query,
            )
            response = self.call_llm_with_prompt(prompt)
            self.logger.log(
                f"""Rephrased Response:  {response}
                """
            )
            return response
        except Exception as exception:
            return (
                "Unfortunately, I was not able to rephrase query, "
                "because of the following error:\n"
                f"\n{exception}\n"
            )

    @property
    def logs(self):
        return self.logger.logs

    @property
    def last_error(self):
        raise NotImplementedError

    @property
    def last_query_log_id(self):
        raise NotImplementedError



"""
Response Parsers for the user to customize response returned from the chat method
"""
from .context import Context
from .response_parser import IResponseParser, ResponseParser
from .streamlit_response import StreamlitResponse

__all__ = ["IResponseParser", "ResponseParser", "StreamlitResponse", "Context"]



from typing import Any

from pandasai.exceptions import LLMNotFoundError
from pandasai.pipelines.base_logic_unit import BaseLogicUnit
from pandasai.prompts.file_based_prompt import FileBasedPrompt


class PromptExecution(BaseLogicUnit):
    def execute(self, input: FileBasedPrompt, **kwargs) -> Any:
        config = kwargs.get("config")
        if config is None or getattr(config, "llm", None) is None:
            raise LLMNotFoundError()
        llm = getattr(config, "llm")
        return llm.call(input)



import re
from abc import ABC, abstractmethod
from decimal import Decimal
from typing import Any, Iterable

from ..df_info import df_type


class BaseOutputType(ABC):
    @property
    @abstractmethod
    def template_hint(self) -> str:
        ...

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    def _validate_type(self, actual_type: str) -> bool:
        return actual_type == self.name

    @abstractmethod
    def _validate_value(self, actual_value):
        ...

    def validate(self, result: dict[str, Any]) -> tuple[bool, Iterable[str]]:
        """
        Validate 'type' and 'value' from the result dict.

        Args:
            result (dict[str, Any]): The result of code execution in
                dict representation. Should have the following schema:
                {
                    "type": <output_type_name>,
                    "value": <generated_value>
                }

        Returns:
             (tuple(bool, Iterable(str)):
                Boolean value whether the result matches output type
                and collection of logs containing messages about
                'type' or 'value' mismatches.
        """
        validation_logs = []
        actual_type, actual_value = result.get("type"), result.get("value")

        type_ok = self._validate_type(actual_type)
        if not type_ok:
            validation_logs.append(
                f"The result dict contains inappropriate 'type'. "
                f"Expected '{self.name}', actual '{actual_type}'."
            )
        value_ok = self._validate_value(actual_value)
        if not value_ok:
            validation_logs.append(
                f"Actual value {repr(actual_value)} seems to be inappropriate "
                f"for the type '{self.name}'."
            )

        return all((type_ok, value_ok)), validation_logs


class NumberOutputType(BaseOutputType):
    @property
    def template_hint(self):
        return """type (must be "number"), value must int. Example: { "type": "number", "value": 125 }"""  # noqa E501

    @property
    def name(self):
        return "number"

    def _validate_value(self, actual_value: Any) -> bool:
        return isinstance(actual_value, (int, float, Decimal))


class DataFrameOutputType(BaseOutputType):
    @property
    def template_hint(self):
        return """type (must be "dataframe"), value must be pd.DataFrame or pd.Series. Example: { "type": "dataframe", "value": pd.DataFrame({...}) }"""  # noqa E501

    @property
    def name(self):
        return "dataframe"

    def _validate_value(self, actual_value: Any) -> bool:
        return bool(df_type(actual_value))


class PlotOutputType(BaseOutputType):
    @property
    def template_hint(self):
        return """type (must be "plot"), value must be string. Example: { "type": "plot", "value": "temp_chart.png" }"""  # noqa E501

    @property
    def name(self):
        return "plot"

    def _validate_value(self, actual_value: Any) -> bool:
        if not isinstance(actual_value, str):
            return False

        path_to_plot_pattern = r"^(\/[\w.-]+)+(/[\w.-]+)*$|^[^\s/]+(/[\w.-]+)*$"
        return bool(re.match(path_to_plot_pattern, actual_value))


class StringOutputType(BaseOutputType):
    @property
    def template_hint(self):
        return """type (must be "string"), value must be string. Example: { "type": "string", "value": f"The highest salary is {highest_salary}." }"""  # noqa E501

    @property
    def name(self):
        return "string"

    def _validate_value(self, actual_value: Any) -> bool:
        return isinstance(actual_value, str)


class DefaultOutputType(BaseOutputType):
    default_types = ["string", "number", "dataframe", "plot"]

    @property
    def template_hint(self):
        return """type (possible values "string", "number", "dataframe", "plot"). Examples: { "type": "string", "value": f"The highest salary is {highest_salary}." } or { "type": "number", "value": 125 } or { "type": "dataframe", "value": pd.DataFrame({...}) } or { "type": "plot", "value": "temp_chart.png" }"""  # noqa E501

    @property
    def name(self):
        return "default"

    def _validate_type(self, actual_type: str) -> bool:
        return True

    def _validate_value(self, actual_value: Any) -> bool:
        return True

    def validate(self, result: dict[str, Any]) -> tuple[bool, Iterable]:
        """
        Validate 'type' and 'value' from the result dict.

        Returns:
             (bool): True since the `DefaultOutputType`
                is supposed to have no validation
        """
        return result["type"] in self.default_types, ()



from ..prompts import BasePrompt


class Callbacks:
    def __init__(self, agent):
        self.agent = agent

    def on_prompt_generation(self, prompt: BasePrompt) -> str:
        """
        A method to be called after prompt generation.

        Args:
            prompt (str): A prompt
        """
        self.agent.last_prompt = str(prompt)

    def on_code_generation(self, code: str):
        """
        A method to be called after code generation.

        Args:
            code (str): A python code
        """
        self.agent.last_code_generated = code

    def before_code_execution(self, code: str):
        """
        A method to be called after code execution.

        Args:
            code (str): A python code
        """
        self.agent.last_code_executed = code

    def on_result(self, result):
        """
        A method to be called after code execution.

        Args:
            result (Any): A python code
        """
        self.agent.last_result = result



import logging
from typing import Any

from pandasai.helpers.logger import Logger
from pandasai.pipelines.logic_unit_output import LogicUnitOutput

from ...helpers.output_validator import OutputValidator
from ..base_logic_unit import BaseLogicUnit
from ..pipeline_context import PipelineContext


class ResultValidation(BaseLogicUnit):
    """
    Result Validation Stage
    """

    pass

    def execute(self, input: Any, **kwargs) -> Any:
        """
        This method will return output according to
        Implementation.

        :param input: Your input data.
        :param kwargs: A dictionary of keyword arguments.
            - 'logger' (any): The logger for logging.
            - 'config' (Config): Global configurations for the test
            - 'context' (any): The execution context.

        :return: The result of the execution.
        """
        pipeline_context: PipelineContext = kwargs.get("context")
        logger: Logger = kwargs.get("logger")

        result = input
        success = False
        message = None
        if result is not None:
            if isinstance(result, dict):
                (
                    validation_ok,
                    validation_logs,
                ) = OutputValidator.validate(
                    pipeline_context.get("output_type"), result
                )
                if not validation_ok:
                    logger.log("\n".join(validation_logs), level=logging.WARNING)
                    success = False
                    message = "Output Validation Failed"

                else:
                    success = True
                    message = "Output Validation Successful"

            pipeline_context.add("last_result", result)
            logger.log(f"Answer: {result}")

        return LogicUnitOutput(result, success, message)



from typing import Any, Union

from pandasai.pipelines.logic_unit_output import LogicUnitOutput

from ...helpers.logger import Logger
from ...prompts.base import BasePrompt
from ...prompts.generate_python_code import GeneratePythonCodePrompt
from ...prompts.generate_python_code_with_sql import GeneratePythonCodeWithSQLPrompt
from ..base_logic_unit import BaseLogicUnit
from ..pipeline_context import PipelineContext


class PromptGeneration(BaseLogicUnit):
    """
    Code Prompt Generation Stage
    """

    pass

    def execute(self, input: Any, **kwargs) -> Any:
        """
        This method will return output according to
        Implementation.

        :param input: Your input data.
        :param kwargs: A dictionary of keyword arguments.
            - 'logger' (any): The logger for logging.
            - 'config' (Config): Global configurations for the test
            - 'context' (any): The execution context.

        :return: The result of the execution.
        """
        self.context: PipelineContext = kwargs.get("context")
        self.logger: Logger = kwargs.get("logger")

        prompt = self.get_chat_prompt(self.context)
        self.logger.log(f"Using prompt: {prompt}")

        return LogicUnitOutput(
            prompt,
            True,
            "Prompt Generated Successfully",
            {"content_type": "prompt", "value": prompt.to_string()},
        )

    def get_chat_prompt(self, context: PipelineContext) -> Union[str, BasePrompt]:
        # set matplotlib as the default library
        viz_lib = "matplotlib"
        if context.config.data_viz_library:
            viz_lib = context.config.data_viz_library

        output_type = context.get("output_type")

        return (
            GeneratePythonCodeWithSQLPrompt(
                context=context,
                last_code_generated=context.get("last_code_generated"),
                viz_lib=viz_lib,
                output_type=output_type,
            )
            if context.config.direct_sql
            else GeneratePythonCodePrompt(
                context=context,
                last_code_generated=context.get("last_code_generated"),
                viz_lib=viz_lib,
                output_type=output_type,
            )
        )



import hashlib
import os
import time
from typing import Optional, Union

import pandasai.pandas as pd

from ..constants import DEFAULT_FILE_PERMISSIONS
from ..helpers.path import find_project_root
from .base import BaseConnector, BaseConnectorConfig


class YahooFinanceConnectorConfig(BaseConnectorConfig):
    """
    Connector configuration for Yahoo Finance.
    """

    dialect: str = "yahoo_finance"
    host: str = "yahoo.finance.com"
    database: str = "stock_data"
    host: str


class YahooFinanceConnector(BaseConnector):
    """
    Yahoo Finance connector for retrieving stock data.
    """

    _cache_interval: int = 600  # 10 minutes

    def __init__(
        self,
        stock_ticker: Optional[str] = None,
        config: Optional[Union[YahooFinanceConnectorConfig, dict]] = None,
        cache_interval: int = 600,
        **kwargs,
    ):
        if not stock_ticker and not config:
            raise ValueError(
                "You must specify either a stock ticker or a config object."
            )

        try:
            import yfinance
        except ImportError as e:
            raise ImportError(
                "Could not import yfinance python package. "
                "Please install it with `pip install yfinance`."
            ) from e

        if not isinstance(config, YahooFinanceConnectorConfig):
            if not config:
                config = {}

            if stock_ticker:
                config["table"] = stock_ticker

            yahoo_finance_config = YahooFinanceConnectorConfig(**config)
        else:
            yahoo_finance_config = config

        self._cache_interval = cache_interval
        super().__init__(yahoo_finance_config)
        self.ticker = yfinance.Ticker(self.config.table)

    def head(self, n: int = 5) -> pd.DataFrame:
        """
        Return the head of the data source that the connector is connected to.

        Returns:
            DataFrame: The head of the data source that the connector is connected to.
            connected to.
        """
        return self.ticker.history(period=f"{n}d")

    def _get_cache_path(self, include_additional_filters: bool = False):
        """
        Return the path of the cache file.

        Returns:
            str: The path of the cache file.
        """
        cache_dir = os.path.join(os.getcwd(), "")
        try:
            cache_dir = os.path.join((find_project_root()), "cache")
        except ValueError:
            cache_dir = os.path.join(os.getcwd(), "cache")

        return os.path.join(cache_dir, f"{self.config.table}_data.parquet")

    def _get_cache_path(self):
        """
        Return the path of the cache file for Yahoo Finance data.
        """
        try:
            cache_dir = os.path.join((find_project_root()), "cache")
        except ValueError:
            cache_dir = os.path.join(os.getcwd(), "cache")

        os.makedirs(cache_dir, mode=DEFAULT_FILE_PERMISSIONS, exist_ok=True)

        return os.path.join(cache_dir, f"{self.config.table}_data.parquet")

    def _cached(self):
        """
        Return the cached Yahoo Finance data if it exists and is not older than the
        cache interval.

        Returns:
            DataFrame|None: The cached data if it exists and is not older than the cache
            interval, None otherwise.
        """
        cache_path = self._get_cache_path()
        if not os.path.exists(cache_path):
            return None

        # If the file is older than 1 day, delete it
        if os.path.getmtime(cache_path) < time.time() - self._cache_interval:
            if self.logger:
                self.logger.log(f"Deleting expired cached data from {cache_path}")
            os.remove(cache_path)
            return None

        if self.logger:
            self.logger.log(f"Loading cached data from {cache_path}")

        return cache_path

    def execute(self):
        """
        Execute the connector and return the result.

        Returns:
            DataFrame: The result of the connector.
        """
        if cached_path := self._cached():
            return pd.read_parquet(cached_path)

        # Use yfinance to retrieve historical stock data
        stock_data = self.ticker.history(period="max")

        # Save the result to the cache
        stock_data.to_parquet(self._get_cache_path())

        return stock_data

    @property
    def rows_count(self):
        """
        Return the number of rows in the data source that the connector is
        connected to.

        Returns:
            int: The number of rows in the data source that the connector is
            connected to.
        """
        stock_data = self.execute()
        return len(stock_data)

    @property
    def columns_count(self):
        """
        Return the number of columns in the data source that the connector is
        connected to.

        Returns:
            int: The number of columns in the data source that the connector is
            connected to.
        """
        stock_data = self.execute()
        return len(stock_data.columns)

    @property
    def column_hash(self):
        """
        Return the hash code that is unique to the columns of the data source
        that the connector is connected to.

        Returns:
            int: The hash code that is unique to the columns of the data source
            that the connector is connected to.
        """
        stock_data = self.execute()
        columns_str = "|".join(stock_data.columns)
        return hashlib.sha256(columns_str.encode("utf-8")).hexdigest()

    @property
    def fallback_name(self):
        """
        Return the fallback name of the connector.

        Returns:
            str: The fallback name of the connector.
        """
        return self.config.table

    @property
    def pandas_df(self):
        return self.execute()



import json
from typing import Any

from pandasai.ee.agents.semantic_agent.prompts.semantic_agent_prompt import (
    SemanticAgentPrompt,
)
from pandasai.helpers.logger import Logger
from pandasai.pipelines.base_logic_unit import BaseLogicUnit
from pandasai.pipelines.logic_unit_output import LogicUnitOutput
from pandasai.pipelines.pipeline_context import PipelineContext


class SemanticPromptGeneration(BaseLogicUnit):
    """
    Code Prompt Generation Stage
    """

    pass

    def execute(self, input: Any, **kwargs) -> Any:
        """
        This method will return output according to
        Implementation.

        :param input: Last logic unit output
        :param kwargs: A dictionary of keyword arguments.
            - 'logger' (any): The logger for logging.
            - 'config' (Config): Global configurations for the test
            - 'context' (any): The execution context.

        :return: LogicUnitOutput(prompt)
        """
        self.context: PipelineContext = kwargs.get("context")
        self.logger: Logger = kwargs.get("logger")

        prompt = SemanticAgentPrompt(
            context=self.context, schema=json.dumps(self.context.get("df_schema"))
        )
        self.logger.log(f"Using prompt: {prompt}")

        return LogicUnitOutput(
            prompt,
            True,
            "Prompt Generated Successfully",
            {"content_type": "prompt", "value": prompt.to_string()},
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional

from ..helpers import load_dotenv
from ..prompts.base import BasePrompt
from .base import LLM

if TYPE_CHECKING:
    from pandasai.pipelines.pipeline_context import PipelineContext


load_dotenv()


class HuggingFaceTextGen(LLM):
    """HuggingFace Text Generation Inference LLM
       Generates text using HuggingFace inference API.

    Attributes:
        max_new_tokens: Max number of tokens to generate.
        top_k: Sample from top k tokens.
        top_p: Sample from top p probability.
        typical_p: typical probability of a token.
        temperature: Controls randomness in the model. Lower values will make the model more deterministic and higher values will make the model more random.
        repetition_penalty: controls the likelihood of repeating same tokens based on value.
        truncate: truncate the input to the maximum length of the model.
        stop_sequences: A stop sequence is a string that stops the model from generating tokens.
        seed: The seed to use for random generation.
        do_sample: Whether or not to use sampling.
        timeout: adding timeout restricts huggingface from waiting indefinitely for model's response.
    """

    max_new_tokens: int = 1024
    top_k: Optional[int] = None
    top_p: Optional[float] = 0.8
    typical_p: Optional[float] = 0.8
    temperature: float = 1e-3  # must be strictly positive
    repetition_penalty: Optional[float] = None
    truncate: Optional[int] = None
    stop_sequences: List[str] = []
    seed: Optional[int] = None
    do_sample: Optional[bool] = False
    inference_server_url: str = ""
    streaming: Optional[bool] = False
    timeout: int = 120
    client: Any

    def __init__(self, inference_server_url: str, **kwargs):
        try:
            import text_generation

            for key, val in kwargs.items():
                if key in self.__annotations__:
                    setattr(self, key, val)

            self.client = text_generation.Client(
                base_url=inference_server_url,
                timeout=self.timeout,
            )

        except ImportError as e:
            raise ImportError(
                "Could not import text_generation python package. "
                "Please install it with `pip install text_generation`."
            ) from e

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling text generation inference API."""
        return {
            "max_new_tokens": self.max_new_tokens,
            "top_k": self.top_k,
            "top_p": self.top_p,
            "typical_p": self.typical_p,
            "temperature": self.temperature,
            "repetition_penalty": self.repetition_penalty,
            "truncate": self.truncate,
            "stop_sequences": self.stop_sequences,
            "do_sample": self.do_sample,
            "seed": self.seed,
        }

    def call(self, instruction: BasePrompt, context: PipelineContext = None) -> str:
        prompt = instruction.to_string()

        memory = context.memory if context else None

        prompt = self.prepend_system_prompt(prompt, memory)

        params = self._default_params
        if self.streaming:
            return "".join(
                chunk.template
                for chunk in self.client.generate_stream(prompt, **params)
            )
        res = self.client.generate(prompt, **params)
        if self.stop_sequences:
            # remove stop sequences from the end of the generated text
            for stop_seq in self.stop_sequences:
                if stop_seq in res.generated_text:
                    res.generated_text = res.generated_text[
                        : res.generated_text.index(stop_seq)
                    ]
        self.last_prompt = prompt
        return res.generated_text

    @property
    def type(self) -> str:
        return "huggingface-text-generation"



from typing import Optional

from pandasai.ee.agents.semantic_agent.pipeline.code_generator import CodeGenerator
from pandasai.ee.agents.semantic_agent.pipeline.error_correction_pipeline.fix_semantic_json_pipeline import (
    FixSemanticJsonPipeline,
)
from pandasai.ee.agents.semantic_agent.pipeline.llm_call import LLMCall
from pandasai.ee.agents.semantic_agent.pipeline.Semantic_prompt_generation import (
    SemanticPromptGeneration,
)
from pandasai.helpers.logger import Logger
from pandasai.helpers.query_exec_tracker import QueryExecTracker
from pandasai.pipelines.chat.code_cleaning import CodeCleaning
from pandasai.pipelines.chat.error_correction_pipeline.error_correction_pipeline_input import (
    ErrorCorrectionPipelineInput,
)
from pandasai.pipelines.pipeline import Pipeline
from pandasai.pipelines.pipeline_context import PipelineContext


class ErrorCorrectionPipeline:
    """
    Error Correction Pipeline to regenerate prompt and code
    """

    _context: PipelineContext
    _logger: Logger

    def __init__(
        self,
        context: Optional[PipelineContext] = None,
        logger: Optional[Logger] = None,
        query_exec_tracker: QueryExecTracker = None,
        on_prompt_generation=None,
        on_code_generation=None,
    ):
        self.pipeline = Pipeline(
            context=context,
            logger=logger,
            query_exec_tracker=query_exec_tracker,
            steps=[
                SemanticPromptGeneration(
                    on_execution=on_prompt_generation,
                ),
                LLMCall(),
                CodeGenerator(
                    on_execution=on_code_generation,
                    on_failure=self.on_wrong_semantic_json,
                ),
                CodeCleaning(),
            ],
        )

        self.fix_semantic_json_pipeline = FixSemanticJsonPipeline(
            context=context,
            logger=logger,
            query_exec_tracker=query_exec_tracker,
            on_code_generation=on_code_generation,
            on_prompt_generation=on_prompt_generation,
        )
        self.query_exec_tracker = query_exec_tracker

        self._context = context
        self._logger = logger

    def run(self, input: ErrorCorrectionPipelineInput):
        self._logger.log(f"Executing Pipeline: {self.__class__.__name__}")
        return self.pipeline.run(input)

    def on_wrong_semantic_json(self, code, errors):
        self.query_exec_tracker.add_step(
            {
                "type": "CodeGenerator",
                "success": False,
                "message": "Failed to validate json",
                "execution_time": None,
                "data": {
                    "content_type": "code",
                    "value": code,
                    "exception": errors,
                },
            }
        )
        correction_input = ErrorCorrectionPipelineInput(code, errors)
        return self.fix_semantic_json_pipeline.run(correction_input)



from __future__ import annotations

from typing import TYPE_CHECKING

from openai import OpenAI

from ..helpers.memory import Memory
from ..prompts.base import BasePrompt
from .base import LLM

if TYPE_CHECKING:
    from pandasai.pipelines.pipeline_context import PipelineContext


class LocalLLM(LLM):
    def __init__(self, api_base: str, model: str = "", api_key: str = "", **kwargs):
        if not api_key:
            api_key = "dummy"

        self.model = model
        self.client = OpenAI(base_url=api_base, api_key=api_key).chat.completions
        self._invocation_params = kwargs

    def chat_completion(self, value: str, memory: Memory) -> str:
        messages = memory.to_openai_messages() if memory else []

        # adding current prompt as latest query message
        messages.append(
            {
                "role": "user",
                "content": value,
            }
        )

        params = {"model": self.model, "messages": messages, **self._invocation_params}
        response = self.client.create(**params)

        return response.choices[0].message.content

    def call(self, instruction: BasePrompt, context: PipelineContext = None) -> str:
        self.last_prompt = instruction.to_string()

        memory = context.memory if context else None

        return self.chat_completion(self.last_prompt, memory)

    @property
    def type(self) -> str:
        return "local"



"""
Databricks Connector to connects you to your Databricks SQL Warhouse on
Azure, AWS and GCP
"""

from typing import Union

from sqlalchemy import create_engine

from ...connectors.base import BaseConnectorConfig
from ...connectors.sql import SQLBaseConnectorConfig, SQLConnector


class DatabricksConnectorConfig(SQLBaseConnectorConfig):
    """
    Connector configuration for DataBricks.
    """

    host: str
    port: int
    token: str
    httpPath: str


class DatabricksConnector(SQLConnector):
    """
    Databricks connectors are used to connect to Databricks Data Cloud.
    """

    def __init__(
        self,
        config: Union[DatabricksConnectorConfig, dict],
        **kwargs,
    ):
        """
        Initialize the Databricks connector with the given configuration.

        Args:
            config (ConnectorConfig): The configuration for the Databricks connector.
        """
        config["dialect"] = "databricks"
        if isinstance(config, dict):
            env_vars = {
                "token": "DATABRICKS_TOKEN",
                "database": "DATABRICKS_DATABASE",
                "host": "DATABRICKS_HOST",
                "port": "DATABRICKS_PORT",
                "httpPath": "DATABRICKS_HTTP_PATH",
            }
            config = self._populate_config_from_env(config, env_vars)

        super().__init__(config, **kwargs)

    def _load_connector_config(self, config: Union[BaseConnectorConfig, dict]):
        return DatabricksConnectorConfig(**config)

    def _init_connection(self, config: DatabricksConnectorConfig):
        """
        Initialize Database Connection

        Args:
            config (DatabricksConnectorConfig): Configurations to load database

        """
        self._engine = create_engine(
            f"{config.dialect}://token:{config.token}@{config.host}:{config.port}?http_path={config.httpPath}"
        )

        self._connection = self._engine.connect()

    def __repr__(self):
        """
        Return the string representation of the Databricks connector.

        Returns:
        str: The string representation of the Databricks connector.
        """
        return (
            f"<{self.__class__.__name__} dialect={self.config.dialect} "
            f"host={self.config.host} port={self.config.port} "
            f"database={self.config.database} httpPath={str(self.config.httpPath)}"
        )

    def equals(self, other):
        if isinstance(other, self.__class__):
            return (
                self.config.dialect,
                self.config.host,
                self.config.port,
                self.config.httpPath,
            ) == (
                other.config.dialect,
                other.config.host,
                other.config.port,
                other.config.httpPath,
            )
        return False



import re

import requests

import pandasai.pandas as pd


def get_google_sheet(src) -> list:
    """
    Returns a 2D array of the contents of the Google Sheet at the given URL

    Args:
        src (str): The URL of the Google Sheet.

    Returns:
        list: A 2D array representing content of the Google Sheet.
    """

    try:
        from bs4 import BeautifulSoup
    except ImportError as e:
        raise ImportError(
            "Could not import beautifulsoup4 python package. "
            "Please install it with `pip install beautifulsoup4`."
        ) from e

    # The size of the Google sheet that can be read is limited
    raw_html = requests.get(src).text

    soup = BeautifulSoup(raw_html, "html.parser")
    table = soup.find("tbody")
    rows = table.find_all("tr")
    grid = []
    for row in rows:
        cols = row.find_all("td")
        clean_row = [col.text for col in cols]
        grid.append(clean_row)
    return grid


def sheet_to_df(sheet) -> list:
    """
    Returns a list of dataframes for each data table in a given spreadsheet

    Args:
        sheet (list): A 2D array of the contents of the Google Sheet

    Returns:
        list: A list of dataframes from the Google Sheet.
    """

    # A dataframe starts when a header is found
    # A header is the first instance of a set of contiguous alphanumeric columns
    # A dataframe ends when a blank row is found or an empty column is found

    num = 0  # The number of the dataframe
    headers = []  # Each header is a tuple (num, row, col_start, col_end)
    binding_headers = []
    dfs = []  # Each df is a tuple (num, df)

    # First pass: get all the headers
    for row in range(len(sheet)):
        # if every cell in the row is empty, skip row
        if all(sheet[row][col].strip() == "" for col in range(len(sheet[row]))):
            headers += binding_headers
            binding_headers = []
            continue

        for col in range(len(sheet[row])):
            # Check if the cell is bounded by a header
            if any(col >= header[2] and col <= header[3] for header in binding_headers):
                continue

            # Check if the cell is commented out
            if sheet[row][col].strip().startswith("//"):
                continue

            if re.search("[a-zA-Z]", sheet[row][col]):
                head_start = col
                head_end = col
                while head_end < len(sheet[row]) and re.search(
                    "[a-zA-Z]", sheet[row][head_end]
                ):
                    head_end += 1
                binding_headers.append([num, row, head_start, head_end])
                num += 1
    headers += binding_headers

    # Second pass: get all the dataframes
    for header in headers:
        df = []
        for row in range(header[1], len(sheet)):
            if all(
                sheet[row][col].strip() == "" for col in range(header[2], header[3])
            ):
                break
            df_row = [sheet[row][col] for col in range(header[2], header[3])]
            df.append(df_row)
        cols = df[0]
        data = df[1:]
        df = pd.DataFrame(data, columns=cols)

        # Cast all the numeric columns to numeric types
        for col in df.columns:
            try:
                df[col] = pd.to_numeric(df[col])
            except ValueError:
                pass
        dfs.append(df)

    return dfs


def from_google_sheets(url) -> list:
    """
    Returns the dataframes that are in a Google sheet.

    Args:
        url (str): The URL of the Google Sheet
    Returns:
        list: A list of dataframes from the Google Sheet.
    """

    sheet = get_google_sheet(url)
    return sheet_to_df(sheet)



from typing import Any, List

from pandasai.connectors.pandas import PandasConnector
from pandasai.connectors.sql import SQLConnector
from pandasai.exceptions import InvalidConfigError
from pandasai.pipelines.logic_unit_output import LogicUnitOutput

from ...connectors import BaseConnector
from ..base_logic_unit import BaseLogicUnit
from ..pipeline_context import PipelineContext


class ValidatePipelineInput(BaseLogicUnit):
    """
    Validates pipeline input
    """

    pass

    def _validate_direct_sql(self, dfs: List[BaseConnector]) -> bool:
        """
        Raises error if they don't belong sqlconnector or have different credentials
        Args:
            dfs (List[BaseConnector]): list of BaseConnectors

        Raises:
            InvalidConfigError: Raise Error in case of config is set but criteria is not met
        """

        if self.context.config.direct_sql:
            if all(
                (isinstance(df, SQLConnector) and df.equals(dfs[0])) for df in dfs
            ) or all(
                (isinstance(df, PandasConnector) and df.sql_enabled) for df in dfs
            ):
                return True
            else:
                raise InvalidConfigError(
                    "Direct requires all Connector and they belong to same datasource "
                    "and have same credentials"
                )
        return False

    def execute(self, input: Any, **kwargs) -> Any:
        """
        This method validates pipeline context and configs

        :param input: Your input data.
        :param kwargs: A dictionary of keyword arguments.
            - 'logger' (any): The logger for logging.
            - 'config' (Config): Global configurations for the test
            - 'context' (any): The execution context.

        :return: The result of the execution.
        """
        self.context: PipelineContext = kwargs.get("context")
        self._validate_direct_sql(self.context.dfs)
        return LogicUnitOutput(input, True, "Input Validation Successful")



from typing import Optional

from pandasai.skills import Skill


class SkillsManager:
    """
    Manages Custom added Skills and tracks used skills for the query
    """

    def __init__(self) -> None:
        self.skills = []
        self.used_skills = []

    def add_skills(self, *skills: Skill):
        """
        Add skills to the list of skills. If a skill with the same name
             already exists, raise an error.

        Args:
            *skills: Variable number of skill objects to add.
        """
        for skill in skills:
            if any(existing_skill.name == skill.name for existing_skill in self.skills):
                raise ValueError(f"Skill with name '{skill.name}' already exists.")

        self.skills.extend(skills)

    def skill_exists(self, name: str):
        """
        Check if a skill with the given name exists in the list of skills.

        Args:
            name (str): The name of the skill to check.

        Returns:
            bool: True if a skill with the given name exists, False otherwise.
        """
        return any(skill.name == name for skill in self.skills)

    def has_skills(self):
        """
        Check if there are any skills in the list of skills.

        Returns:
            bool: True if there are skills, False otherwise.
        """
        return len(self.skills) > 0

    def get_skill_by_func_name(self, name: str):
        """
        Get a skill by its name.

        Args:
            name (str): The name of the skill to retrieve.

        Returns:
            Skill or None: The skill with the given name, or None if not found.
        """
        return next((skill for skill in self.skills if skill.name == name), None)

    def add_used_skill(self, skill: str):
        if self.skill_exists(skill):
            self.used_skills.append(skill)

    def __str__(self) -> str:
        """
        Present all skills
        Returns:
            str: _description_
        """
        return "".join(str(skill) for skill in self.skills)

    def prompt_display(self) -> Optional[str]:
        """
        Displays skills for prompt
        """
        if len(self.skills) == 0:
            return None

        return f"You are already provided with the following functions that you can call:\n{self}"

    def to_object(self) -> str:
        return [skill.stringify() for skill in self.skills]



import re


def extract_table_names(sql_query):
    # Regular expression pattern to match table names in FROM and JOIN clauses
    pattern = r"(?:FROM|JOIN)\s+(?<!:)(\"?\w+\"?)(?=\s|$)"

    # Find all matches in the SQL query
    matches = re.findall(pattern, sql_query, re.IGNORECASE)
    return matches



from .base import BasePrompt


class GeneratePythonCodePrompt(BasePrompt):
    """Prompt to generate Python code from a dataframe."""

    template_path = "generate_python_code.tmpl"

    def to_json(self):
        context = self.props["context"]
        viz_lib = self.props["viz_lib"]
        output_type = self.props["output_type"]
        memory = context.memory
        conversations = memory.to_json()

        system_prompt = memory.get_system_prompt()

        # prepare datasets
        datasets = [dataset.to_json() for dataset in context.dfs]

        return {
            "datasets": datasets,
            "conversation": conversations,
            "system_prompt": system_prompt,
            "prompt": self.to_string(),
            "config": {
                "direct_sql": context.config.direct_sql,
                "viz_lib": viz_lib,
                "output_type": output_type,
            },
        }



"""
Base connector class to be extended by all connectors.
"""

import json
import os
from abc import ABC, abstractmethod
from functools import cache
from typing import TYPE_CHECKING, List, Optional, Union

import pandasai.pandas as pd
from pandasai.helpers.dataframe_serializer import (
    DataframeSerializer,
    DataframeSerializerType,
)
from pandasai.pydantic import BaseModel

from ..helpers.logger import Logger

if TYPE_CHECKING:
    from pandasai.ee.connectors.relations import AbstractRelation


class BaseConnectorConfig(BaseModel):
    """
    Base Connector configuration.
    """

    database: str
    table: str
    where: list[list[str]] = None
    connect_args: Optional[dict] = {}


class BaseConnector(ABC):
    """
    Base connector class to be extended by all connectors.
    """

    _logger: Logger = None
    _additional_filters: list[list[str]] = None

    def __init__(
        self,
        config: Union[BaseConnectorConfig, dict],
        name: str = None,
        description: str = None,
        custom_head: pd.DataFrame = None,
        field_descriptions: dict = None,
        connector_relations: List["AbstractRelation"] = None,
    ):
        """
        Initialize the connector with the given configuration.

        Args:
            config (dict): The configuration for the connector.
        """
        if isinstance(config, dict):
            config = self._load_connector_config(config)

        self.config = config
        self.name = name
        self.description = description
        self.custom_head = custom_head
        self.field_descriptions = field_descriptions
        self.connector_relations = connector_relations

    def _load_connector_config(self, config: Union[BaseConnectorConfig, dict]):
        """Loads passed Configuration to object

        Args:
            config (BaseConnectorConfig): Construct config in structure

        Returns:
            config: BaseConnectorConfig
        """
        pass

    def _populate_config_from_env(self, config: dict, envs_mapping: dict):
        """
        Populate the configuration dictionary with values from environment variables
        if not exists in the config.

        Args:
            config (dict): The configuration dictionary to be populated.
            envs_mapping (dict): The dictionary representing a map of config's keys
                and according names of the environment variables.

        Returns:
            dict: The populated configuration dictionary.
        """

        for key, env_var in envs_mapping.items():
            if key not in config and os.getenv(env_var):
                config[key] = os.getenv(env_var)

        return config

    def _init_connection(self, config: BaseConnectorConfig):
        """
        make connection to database
        """
        pass

    @abstractmethod
    def head(self, n: int = 3) -> pd.DataFrame:
        """
        Return the head of the data source that the connector is connected to.
        This information is passed to the LLM to provide the schema of the
        data source.
        """
        pass

    @abstractmethod
    def execute(self) -> pd.DataFrame:
        """
        Execute the given query on the data source that the connector is
        connected to.
        """
        pass

    def set_additional_filters(self, filters: dict):
        """
        Add additional filters to the connector.

        Args:
            filters (dict): The additional filters to add to the connector.
        """
        self._additional_filters = filters or []

    @property
    def rows_count(self):
        """
        Return the number of rows in the data source that the connector is
        connected to.
        """
        raise NotImplementedError

    @property
    def columns_count(self):
        """
        Return the number of columns in the data source that the connector is
        connected to.
        """
        raise NotImplementedError

    @property
    def column_hash(self):
        """
        Return the hash code that is unique to the columns of the data source
        that the connector is connected to.
        """
        raise NotImplementedError

    @property
    def path(self):
        """
        Return the path of the data source that the connector is connected to.
        """
        # JDBC string
        path = f"{self.__class__.__name__}://{self.config.host}:"
        if hasattr(self.config, "port"):
            path += str(self.config.port)
        path += f"/{self.config.database}/{self.config.table}"
        return path

    @property
    def logger(self):
        """
        Return the logger for the connector.
        """
        return self._logger

    @logger.setter
    def logger(self, logger: Logger):
        """
        Set the logger for the connector.

        Args:
            logger (Logger): The logger for the connector.
        """
        self._logger = logger

    @property
    def fallback_name(self):
        """
        Return the name of the table that the connector is connected to.
        """
        raise NotImplementedError

    @property
    def pandas_df(self):
        """
        Returns the pandas dataframe
        """
        raise NotImplementedError

    @property
    def type(self):
        return "pd.DataFrame"

    def equals(self, other):
        return self.__dict__ == other.__dict__

    @cache
    def get_head(self, n: int = 3) -> pd.DataFrame:
        """
        Return the head of the data source that the connector is connected to.
        This information is passed to the LLM to provide the schema of the
        data source.

        Args:
            n (int, optional): The number of rows to return. Defaults to 5.

        Returns:
            pd.DataFrame: The head of the data source that the connector is
                connected to.
        """
        return self.custom_head if self.custom_head is not None else self.head(n)

    def head_with_truncate_columns(self, max_size=25) -> pd.DataFrame:
        """
        Truncate the columns of the dataframe to a maximum of 20 characters.

        Args:
            df (pd.DataFrame): The dataframe to truncate the columns of.

        Returns:
            pd.DataFrame: The dataframe with truncated columns.
        """
        df_trunc = self.get_head().copy()

        for col in df_trunc.columns:
            if df_trunc[col].dtype == "object":
                first_val = df_trunc[col].iloc[0]
                if isinstance(first_val, str) and len(first_val) > max_size:
                    df_trunc[col] = f"{df_trunc[col].str.slice(0, max_size - 3)}..."

        return df_trunc

    @cache
    def get_schema(self) -> pd.DataFrame:
        """
        A sample of the dataframe.

        Returns:
            pd.DataFrame: A sample of the dataframe.
        """
        if self.get_head() is None:
            return None

        if len(self.get_head()) > 0:
            return self.head_with_truncate_columns()

        return self.get_head()

    @cache
    def to_csv(self) -> str:
        """
        A proxy-call to the dataframe's `.to_csv()`.

        Returns:
            str: The dataframe as a CSV string.
        """
        return self.get_head().to_csv(index=False)

    @cache
    def to_string(
        self,
        index: int = 0,
        is_direct_sql: bool = False,
        serializer: DataframeSerializerType = None,
        enforce_privacy: bool = False,
    ) -> str:
        """
        Convert dataframe to string
        Returns:
            str: dataframe string
        """
        # If field descriptions are added always use YML. Other formats don't support field descriptions yet
        if self.field_descriptions or self.connector_relations:
            serializer = DataframeSerializerType.YML

        return DataframeSerializer().serialize(
            self,
            extras={
                "index": index,
                "type": "pd.DataFrame",
                "is_direct_sql": is_direct_sql,
                "enforce_privacy": enforce_privacy,
            },
            type_=serializer,
        )

    @cache
    def to_json(self):
        df_head = self.get_head()

        return {
            "name": self.name,
            "description": self.description,
            "head": json.loads(df_head.to_json(orient="records", date_format="iso")),
        }

    def register_enable_sql_query(self, table_name=None):
        raise NotImplementedError



import os
from pathlib import Path

from ..exceptions import TemplateFileNotFoundError
from .base import AbstractPrompt


class FileBasedPrompt(AbstractPrompt):
    """Base class for prompts supposed to read template content from a file.

    `_path_to_template` attribute has to be specified.
    """

    _path_to_template: str

    def __init__(self, **kwargs):
        if (template_path := kwargs.pop("path_to_template", None)) is not None:
            self._path_to_template = template_path
        else:
            current_dir_path = Path(__file__).parent
            self._path_to_template = os.path.join(
                current_dir_path, "..", self._path_to_template
            )

        self.conversation_text = self.template
        super().__init__(**kwargs)

    @property
    def template(self) -> str:
        try:
            with open(self._path_to_template, encoding="utf-8") as fp:
                return fp.read()
        except FileNotFoundError as e:
            raise TemplateFileNotFoundError(
                self._path_to_template, self.__class__.__name__
            ) from e
        except IOError as exc:
            raise RuntimeError(
                f"Failed to read template file '{self._path_to_template}': {exc}"
            ) from exc



"""
Polars connector class to handle csv, parquet, xlsx files and polars dataframes.
"""

from functools import cache, cached_property
from typing import TYPE_CHECKING, Any, Union

import pandasai.pandas as pd

# Use a conditional import for type checking
if TYPE_CHECKING:
    try:
        import polars as pl

        PolarsDataFrame = pl.DataFrame
        PolarsSeries = pl.Series
    except ImportError:
        PolarsDataFrame = Any
        PolarsSeries = Any
else:
    PolarsDataFrame = Any
    PolarsSeries = Any

from pydantic import BaseModel

from ..helpers.data_sampler import DataSampler
from ..helpers.file_importer import FileImporter
from ..helpers.logger import Logger
from .base import BaseConnector


class PolarsConnectorConfig(BaseModel):
    """
    Polars Connector configuration.
    """

    original_df: PolarsDataFrame

    class Config:
        arbitrary_types_allowed = True


class PolarsConnector(BaseConnector):
    """
    Polars connector class to handle csv, parquet, xlsx files and polars dataframes.
    """

    pandas_df = pd.DataFrame
    _logger: Logger = None
    _additional_filters: list[list[str]] = None

    def __init__(
        self,
        config: Union[PolarsConnectorConfig, dict],
        **kwargs,
    ):
        """
        Initialize the Polars connector with the given configuration.

        Args:
            config (PolarsConnectorConfig): The configuration for the Polars connector.
        """

        super().__init__(config, **kwargs)

        self._load_df(self.config.original_df)

    def _load_df(self, df: Union[PolarsDataFrame, PolarsSeries, str, dict]):
        """
        Load the dataframe from a file or polars dataframe.

        Args:
            df (Union[pl.DataFrame, pl.Series, str, dict]): The dataframe to load.
        """
        polars_df = None
        if isinstance(df, pl.Series):
            polars_df = df.to_frame()
        elif isinstance(df, pl.DataFrame):
            polars_df = df
        elif isinstance(df, str):
            polars_df = FileImporter.import_from_file(df)
        elif isinstance(df, dict):
            try:
                polars_df = pl.DataFrame(df)
            except Exception as e:
                raise ValueError(
                    "Invalid input data. We cannot convert it to a dataframe."
                ) from e
        else:
            raise ValueError("Invalid input data. We cannot convert it to a dataframe.")

        self.pandas_df = polars_df.to_pandas()

    def _load_connector_config(
        self, config: Union[PolarsConnectorConfig, dict]
    ) -> PolarsConnectorConfig:
        """
        Loads passed Configuration to object

        Args:
            config (PolarsConnectorConfig): Construct config in structure

            Returns:
                config: PolarsConnectorConfig
        """
        return PolarsConnectorConfig(**config)

    @cache
    def head(self, n: int = 5) -> PolarsDataFrame:
        """
        Return the head of the data source that the connector is connected to.
        This information is passed to the LLM to provide the schema of the
        data source.
        """
        if self.pandas_df is None:
            return None

        sampler = DataSampler(self.pandas_df)
        return sampler.sample(n)

    @cache
    def execute(self) -> PolarsDataFrame:
        """
        Execute the given query on the data source that the connector is
        connected to.
        """
        return self.pandas_df

    @cached_property
    def rows_count(self):
        """
        Return the number of rows in the data source that the connector is
        connected to.
        """
        return len(self.pandas_df) if self.pandas_df.index else 0

    @cached_property
    def columns_count(self):
        """
        Return the number of columns in the data source that the connector is
        connected to.
        """
        return len(self.pandas_df.columns) if self.pandas_df.columns else 0

    @property
    def column_hash(self):
        """
        Return the hash code that is unique to the columns of the data source
        that the connector is connected to.
        """
        import hashlib

        columns_str = "".join(self.pandas_df.columns)
        hash_object = hashlib.sha256(columns_str.encode())
        return hash_object.hexdigest()

    @cached_property
    def path(self):
        """
        Return the path of the data source that the connector is connected to.
        """
        pass

    @property
    def fallback_name(self):
        """
        Return the name of the table that the connector is connected to.
        """
        pass



"""
This module contains helper functions for anonymizing data and generating random data
 before sending it to the LLM (An External API).

Only df.head() is sent to LLM API, hence the df.head() is processed
 to remove any personal or sensitive information.

"""

import random

import numpy as np

import pandasai.pandas as pd

from .anonymizer import Anonymizer


class DataSampler:
    def __init__(self, df: pd.DataFrame):
        """
        Args:
            df (pd.DataFrame): pd.DataFrame to sample from.
        """
        self.df = df

    def sample(self, n: int = 3) -> pd.DataFrame:
        """Sample the dataframe.

        Args:
            n (int, optional): Number of rows to sample. Defaults to 5.

        Returns:
            pd.DataFrame: Sampled dataframe.
        """
        sampled_df = pd.DataFrame()
        if len(self.df) <= n:
            sampled_df = self.df.sample(frac=1)
        else:
            for col in self.df.columns:
                col_sample = self._sample_column(col, n)
                sampled_df[col] = col_sample

        # anonymize the sampled dataframe head
        sampled_df = Anonymizer.anonymize_dataframe_head(sampled_df)

        return sampled_df

    def _sample_column(self, col: str, n: int) -> list:
        """Sample a column.

        Args:
            col (str): Column name.
            n (int): Number of rows to sample.

        Returns:
            list: Sampled column.
        """

        col_sample = []
        col_values = self.df[col].dropna().unique()

        # if there is a null value in the column, it MUST be included in the sample
        if self.df[col].isna().any():
            col_sample.append(np.nan)
            n -= 1

        # if the column has less than n unique values, then a random sample of the
        # unique values should be returned
        if len(col_values) <= n:
            col_sample.extend(col_values)
            n -= len(col_values)
        else:
            col_sample.extend(random.sample(list(col_values), n))
            n = 0

        # if there are still rows to sample, sample them randomly
        if n > 0 and len(col_values) > 0:
            col_sample.extend(random.choices(list(col_values), k=n))
        else:
            col_sample.extend([np.nan] * n)

        # shuffle the column sample before returning it
        random.shuffle(col_sample)
        return col_sample



from typing import List, Optional, Type, Union

import pandas as pd

from pandasai.agent.base import BaseAgent
from pandasai.agent.base_judge import BaseJudge
from pandasai.agent.base_security import BaseSecurity
from pandasai.connectors.base import BaseConnector
from pandasai.pipelines.chat.generate_chat_pipeline import GenerateChatPipeline
from pandasai.schemas.df_config import Config
from pandasai.vectorstores.vectorstore import VectorStore


class Agent(BaseAgent):
    def __init__(
        self,
        dfs: Union[
            pd.DataFrame, BaseConnector, List[Union[pd.DataFrame, BaseConnector]]
        ],
        config: Optional[Union[Config, dict]] = None,
        memory_size: Optional[int] = 10,
        pipeline: Optional[Type[GenerateChatPipeline]] = None,
        vectorstore: Optional[VectorStore] = None,
        description: str = None,
        judge: BaseJudge = None,
        security: BaseSecurity = None,
    ):
        super().__init__(
            dfs, config, memory_size, vectorstore, description, security=security
        )

        self.pipeline = (
            pipeline(
                self.context,
                self.logger,
                on_prompt_generation=self._callbacks.on_prompt_generation,
                on_code_generation=self._callbacks.on_code_generation,
                before_code_execution=self._callbacks.before_code_execution,
                on_result=self._callbacks.on_result,
                judge=judge,
            )
            if pipeline
            else GenerateChatPipeline(
                self.context,
                self.logger,
                on_prompt_generation=self._callbacks.on_prompt_generation,
                on_code_generation=self._callbacks.on_code_generation,
                before_code_execution=self._callbacks.before_code_execution,
                on_result=self._callbacks.on_result,
                judge=judge,
            )
        )

    @property
    def last_error(self):
        return self.pipeline.last_error

    @property
    def last_query_log_id(self):
        return self.pipeline.get_last_track_log_id()



import inspect
from typing import Any, Callable, Optional, Union

from pandasai.pydantic import BaseModel, PrivateAttr


class Skill(BaseModel):
    """Skill that takes a function usable by pandasai"""

    func: Callable[..., Any]
    description: Optional[str] = None
    name: Optional[str] = None
    _signature: Optional[str] = PrivateAttr()

    def __init__(
        self,
        func: Callable[..., Any],
        description: Optional[str] = None,
        name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initializes the skill.

        Args:
            func: The function from which to create a skill
            description: The description of the skill.
                Defaults to the function docstring.
            name: The name of the function. Mandatory when `func` is a lambda.
                Defaults to the functions name.
            **kwargs: additional params
        """

        name = name or func.__name__
        description = description or func.__doc__
        if description is None:
            # if description is None then the function doesn't have a docstring
            # and the user didn't provide any description
            raise ValueError(
                f"Function must have a docstring if no description is provided for skill {name}."
            )
        signature = f"def {name}{inspect.signature(func)}:"

        self._signature = signature
        super(Skill, self).__init__(
            func=func, description=description, name=name, **kwargs
        )

    def __call__(self, *args, **kwargs) -> Any:
        """Calls the skill function"""
        return self.func(*args, **kwargs)

    @classmethod
    def from_function(cls, func: Callable, **kwargs: Any) -> "Skill":
        """
        Creates a skill object from a function

        Args:
            func: The function from which to create a skill

        Returns:
            the `Skill` object

        """
        return cls(func=func, **kwargs)

    def stringify(self):
        return inspect.getsource(self.func)

    def __str__(self):
        return f"""
<function>
{self._signature}
    \"\"\"{self.description}\"\"\"
</function>"""


def skill(*args: Union[str, Callable]) -> Callable:
    """Decorator to create a skill out of functions
    Can be used without arguments. The function must have a docstring.

    Args:
        *args: The arguments to the skill

    Examples:
        .. code-block:: python

            @skill
            def compute_flight_prices(offers: pd.Dataframe) -> List[float]:
                \"\"\"Computes the flight prices\"\"\"
                return

            @skill("custom_name")
            def compute_flight_prices(offers: pd.Dataframe) -> List[float]:
                \"\"\"Computes the flight prices\"\"\"
                return
    """

    def _make_skill_with_name(skill_name: str) -> Callable:
        def _make_skill(skill_fn: Callable) -> Skill:
            return Skill(
                name=skill_name,  # func.__name__ if None
                # when this decorator is used, the function MUST have a docstring
                description=skill_fn.__doc__,
                func=skill_fn,
            )

        return _make_skill

    if len(args) == 1 and isinstance(args[0], str):
        # Example: @skill("skillName")
        return _make_skill_with_name(args[0])
    elif len(args) == 1 and callable(args[0]):
        # Example: @skill
        return _make_skill_with_name(args[0].__name__)(args[0])
    elif not args:
        # Covers the case in which a function is decorated with "@skill()"
        # with the intended behavior of "@skill"
        def _func_wrapper(fn: Callable) -> Skill:
            return _make_skill_with_name(fn.__name__)(fn)

        return _func_wrapper
    else:
        raise ValueError(
            f"Too many arguments for skill decorator. Received: {len(args)}"
        )



import json
from typing import List, Optional, Type, Union

import pandas as pd

from pandasai.agent.base import BaseAgent
from pandasai.agent.base_judge import BaseJudge
from pandasai.connectors.base import BaseConnector
from pandasai.connectors.pandas import PandasConnector
from pandasai.constants import PANDASBI_SETUP_MESSAGE
from pandasai.ee.agents.semantic_agent.pipeline.code_generator import CodeGenerator
from pandasai.ee.agents.semantic_agent.pipeline.semantic_chat_pipeline import (
    SemanticChatPipeline,
)
from pandasai.ee.agents.semantic_agent.prompts.generate_df_schema import (
    GenerateDFSchemaPrompt,
)
from pandasai.ee.helpers.json_helper import extract_json_from_json_str
from pandasai.exceptions import InvalidConfigError, InvalidSchemaJson, InvalidTrainJson
from pandasai.helpers.cache import Cache
from pandasai.helpers.memory import Memory
from pandasai.llm.bamboo_llm import BambooLLM
from pandasai.pipelines.chat.generate_chat_pipeline import GenerateChatPipeline
from pandasai.pipelines.pipeline import Pipeline
from pandasai.pipelines.pipeline_context import PipelineContext
from pandasai.schemas.df_config import Config
from pandasai.vectorstores.vectorstore import VectorStore


class SemanticAgent(BaseAgent):
    """
    Answer Semantic queries
    """

    def __init__(
        self,
        dfs: Union[
            pd.DataFrame, BaseConnector, List[Union[pd.DataFrame, BaseConnector]]
        ],
        config: Optional[Union[Config, dict]] = None,
        schema: Optional[List[dict]] = None,
        memory_size: Optional[int] = 10,
        pipeline: Optional[Type[GenerateChatPipeline]] = None,
        vectorstore: Optional[VectorStore] = None,
        description: str = None,
        judge: BaseJudge = None,
    ):
        super().__init__(dfs, config, memory_size, vectorstore, description)

        self._validate_config()

        self._schema_cache = Cache("schema")
        self._schema = schema or None

        self._create_schema()

        self._sort_dfs_according_to_schema()

        self.init_duckdb_instance()

        # semantic agent works only with direct sql true
        self.config.direct_sql = True

        self.context = PipelineContext(
            dfs=self.dfs,
            config=self.config,
            memory=Memory(memory_size, agent_info=description),
            vectorstore=self._vectorstore,
            initial_values={"df_schema": self._schema},
        )

        self.pipeline = (
            pipeline(
                self.context,
                self.logger,
                judge=judge,
                on_prompt_generation=self._callbacks.on_prompt_generation,
                on_code_generation=self._callbacks.on_code_generation,
                before_code_execution=self._callbacks.before_code_execution,
                on_result=self._callbacks.on_result,
            )
            if pipeline
            else SemanticChatPipeline(
                self.context,
                self.logger,
                judge=judge,
                on_prompt_generation=self._callbacks.on_prompt_generation,
                on_code_generation=self._callbacks.on_code_generation,
                before_code_execution=self._callbacks.before_code_execution,
                on_result=self._callbacks.on_result,
            )
        )

    def validate_and_convert_json(self, jsons):
        json_strs = []

        try:
            for json_data in jsons:
                if isinstance(json_data, str):
                    json.loads(json_data)
                    json_strs.append(json_data)
                elif isinstance(json_data, dict):
                    json_strs.append(json.dumps(json_data))

        except Exception as e:
            raise InvalidTrainJson("Error validating JSON string") from e

        return json_strs

    def train(
        self,
        queries: Optional[List[str]] = None,
        jsons: Optional[List[Union[dict, str]]] = None,
        docs: Optional[List[str]] = None,
    ) -> None:
        json_strs = self.validate_and_convert_json(jsons) if jsons else None

        super().train(queries=queries, codes=json_strs, docs=docs)

    def query(self, query):
        query_pipeline = Pipeline(
            context=self.context,
            logger=self.logger,
            steps=[
                CodeGenerator(),
            ],
        )
        code = query_pipeline.run(query)

        self.execute_code(code)

    def init_duckdb_instance(self):
        for index, tables in enumerate(self._schema):
            if isinstance(self.dfs[index], PandasConnector):
                self._sync_pandas_dataframe_schema(self.dfs[index], tables)
                self.dfs[index].enable_sql_query(tables["table"])

    def _sync_pandas_dataframe_schema(self, df: PandasConnector, schema: dict):
        for dimension in schema["dimensions"]:
            if dimension["type"] in ["date", "datetime", "timestamp"]:
                column = dimension["sql"]
                df.pandas_df[column] = pd.to_datetime(df.pandas_df[column])

    def _sort_dfs_according_to_schema(self):
        schema_dict = {
            table["table"]: [dim["sql"] for dim in table["dimensions"]]
            for table in self._schema
        }
        sorted_dfs = []

        for table in self._schema:
            matched = False
            for df in self.dfs:
                df_columns = df.get_head().columns
                if all(column in df_columns for column in schema_dict[table["table"]]):
                    sorted_dfs.append(df)
                    matched = True

            if not matched:
                raise InvalidSchemaJson(
                    f"Some sql column of table {table['table']} doesn't match with any dataframe"
                )

        self.dfs = sorted_dfs

    def _create_schema(self):
        """
        Generate schema on the initialization of Agent class
        """
        if self._schema:
            self.logger.log(f"using user provided schema: {self._schema}")
            return

        key = self._get_schema_cache_key()
        if self.config.enable_cache:
            value = self._schema_cache.get(key)
            if value is not None:
                self._schema = json.loads(value)
                self.logger.log(f"using schema: {self._schema}")
                return

        prompt = GenerateDFSchemaPrompt(context=self.context)

        result = self.call_llm_with_prompt(prompt)
        self.logger.log(
            f"""Initializing Schema:  {result}
            """
        )
        self._schema = result.replace("# SAMPLE SCHEMA", "")
        schema_data = extract_json_from_json_str(result.replace("# SAMPLE SCHEMA", ""))
        if isinstance(schema_data, dict):
            schema_data = [schema_data]

        self._schema = schema_data
        # save schema in the cache
        if self.config.enable_cache:
            self._schema_cache.set(key, json.dumps(self._schema))

        self.logger.log(f"using schema: {self._schema}")

    def _validate_config(self):
        if not isinstance(self.config.llm, BambooLLM):
            raise InvalidConfigError(
                f"""Semantic Agent works only with BambooLLM follow instructions for setup:\n {PANDASBI_SETUP_MESSAGE}"""
            )

    def _get_schema_cache_key(self):
        """
        Generate schema key
        Returns:
            str: key of schema stored in db
        """
        return "".join(str(df.column_hash) for df in self.context.dfs)

    @property
    def last_error(self):
        return self.pipeline.last_error

    @property
    def last_query_log_id(self):
        return self.pipeline.get_last_track_log_id()



""" Memory class to store the conversations """
from typing import Union


class Memory:
    """Memory class to store the conversations"""

    _messages: list
    _memory_size: int
    _agent_info: str

    def __init__(self, memory_size: int = 1, agent_info: Union[str, None] = None):
        self._messages = []
        self._memory_size = memory_size
        self._agent_info = agent_info

    def add(self, message: str, is_user: bool):
        self._messages.append({"message": message, "is_user": is_user})

    def count(self) -> int:
        return len(self._messages)

    def all(self) -> list:
        return self._messages

    def last(self) -> dict:
        return self._messages[-1]

    def _truncate(self, message: Union[str, int], max_length: int = 100) -> str:
        """
        Truncates the message if it is longer than max_length
        """
        return (
            f"{message[:max_length]} ..." if len(str(message)) > max_length else message
        )

    def get_messages(self, limit: int = None) -> list:
        """
        Returns the conversation messages based on limit parameter
        or default memory size
        """
        limit = self._memory_size if limit is None else limit

        return [
            f"{'### QUERY' if message['is_user'] else '### ANSWER'}\n {message['message'] if message['is_user'] else self._truncate(message['message'])}"
            for message in self._messages[-limit:]
        ]

    def get_conversation(self, limit: int = None) -> str:
        """
        Returns the conversation messages based on limit parameter
        or default memory size
        """
        return "\n".join(self.get_messages(limit))

    def get_previous_conversation(self) -> str:
        """
        Returns the previous conversation but the last message
        """
        messages = self.get_messages(self._memory_size)
        return "" if len(messages) <= 1 else "\n".join(messages[:-1])

    def get_last_message(self) -> str:
        """
        Returns the last message in the conversation
        """
        messages = self.get_messages(self._memory_size)
        return "" if len(messages) == 0 else messages[-1]

    def get_system_prompt(self) -> str:
        return self._agent_info

    def to_json(self):
        messages = []
        for message in self.all():
            if message["is_user"]:
                messages.append({"role": "user", "message": message["message"]})
            else:
                messages.append({"role": "assistant", "message": message["message"]})
        return messages

    def to_openai_messages(self):
        """
        Returns the conversation messages in the format expected by the OpenAI API
        """
        messages = []
        if self.agent_info:
            messages.append(
                {
                    "role": "system",
                    "content": self.get_system_prompt(),
                }
            )
        for message in self.all():
            if message["is_user"]:
                messages.append({"role": "user", "content": message["message"]})
            else:
                messages.append({"role": "assistant", "content": message["message"]})
        return messages

    def clear(self):
        self._messages = []

    @property
    def size(self):
        return self._memory_size

    @property
    def agent_info(self):
        return self._agent_info



import re

from pandasai.exceptions import InvalidSchemaJson

MISSING_TABLE_NAME_MESSAGE = "All measures, dimensions, timeDimensions, order and filters must have the format Table_Name.Dimension or Table_Name.Measure"
TABLE_NOT_FOUND_MESSAGE = "Table {0} Doesn't exist"


class QueryBuilder:
    """
    Creates query from json structure
    """

    def __init__(self, schema):
        self.schema = schema
        self.supported_aggregations = {"sum", "count", "avg", "min", "max"}
        self.supported_granularities = {
            "year",
            "month",
            "day",
            "hour",
            "minute",
            "second",
        }
        self.supported_date_ranges = {
            "last week",
            "last month",
            "this month",
            "this week",
            "today",
            "this year",
            "last year",
        }

    def generate_sql(self, query):
        self._validate_query(query)
        measures = query.get("measures", [])
        dimensions = query.get("dimensions", [])
        time_dimensions = query.get("timeDimensions", [])
        filters = query.get("filters", [])

        columns = self._generate_columns(dimensions, time_dimensions, measures)

        referenced_tables = self._get_referenced_tables(
            dimensions, time_dimensions, measures, filters
        )
        main_table_entry = self._get_main_table_entry(measures, dimensions)

        if not main_table_entry:
            raise ValueError("Table not found in schema.")

        sql = self._build_select_clause(columns)
        sql += self._build_from_clause(main_table_entry)
        sql += self._build_joins_clause(main_table_entry, referenced_tables)
        sql += self._build_where_clause(filters, time_dimensions)
        sql += self._build_group_by_clause(dimensions, time_dimensions)
        sql += self._build_having_clause(filters)
        sql += self._build_order_clause(query)
        sql += self._build_limit_clause(query)

        return sql

    def _validate_table(self, value: str):
        value_splitted = value.split(".")
        if len(value_splitted) == 1:
            raise InvalidSchemaJson(MISSING_TABLE_NAME_MESSAGE)

        table = self.find_table(value_splitted[0])
        if not table:
            raise InvalidSchemaJson(TABLE_NOT_FOUND_MESSAGE.format(value_splitted[0]))

    def _validate_query(self, query: dict):
        for measure in query.get("measures", []):
            self._validate_table(measure)

        for dimension in query.get("dimensions", []):
            self._validate_table(dimension)

        for dimension in query.get("timeDimensions", []):
            self._validate_table(dimension["dimension"])

        for order in query.get("order", []):
            self._validate_table(order["id"])

        for filter in query.get("filters", []):
            self._validate_table(filter["member"])

    def _validate_fix_query(self, query):
        for index, measure in enumerate(query.get("measures", [])):
            query["measures"][index] = self._validate_and_fix_mapped_measure(measure)

        for index, dimension in enumerate(query.get("dimensions", [])):
            query["dimensions"][index] = self._validate_and_fix_mapped_dimension(
                dimension
            )

        for index, dimension in enumerate(query.get("timeDimensions", [])):
            query["timeDimensions"][index][
                "dimension"
            ] = self._validate_and_fix_mapped_dimension(dimension["dimension"])

        for index, order in enumerate(query.get("order", [])):
            query["order"][index]["id"] = self._validate_and_fix_mapped_order(
                order["id"]
            )

        for index, filter in enumerate(query.get("filters", [])):
            query["filters"][index]["member"] = self._validate_and_fix_mapped_order(
                filter["member"]
            )

        return query

    def _generate_columns(self, dimensions, time_dimensions, measures):
        all_dimensions = list(dict.fromkeys(dimensions))
        # + [td["dimension"] for td in time_dimensions]
        columns = []

        for dim in all_dimensions:
            table = self.find_table(dim.split(".")[0])["table"]
            dimension_info = self.find_dimension(dim)
            sql_expr = dimension_info.get("sql")
            name = dimension_info["name"]
            if sql_expr:
                columns.append(f"`{table}`.`{sql_expr}` AS {name}")
            else:
                columns.append(f"{name}")

        for measure in measures:
            table = self.find_table(measure.split(".")[0])["table"]
            measure_info = self.find_measure(measure)
            if measure_info["type"] not in self.supported_aggregations:
                raise ValueError(
                    f"Unsupported aggregation type '{measure_info['type']}' for measure '{measure_info['name']}'. Supported types are: {', '.join(self.supported_aggregations)}"
                )
            sql_expr = measure_info.get("sql") or measure_info["name"]
            columns.append(
                f"{measure_info['type'].upper()}(`{table}`.`{sql_expr}`) AS {measure_info['name']}"
            )

        for time_dimension in time_dimensions:
            columns.append(self._generate_time_dimension_column(time_dimension))

        return list(dict.fromkeys(columns))  # preserve order and return unique columns

    def _validate_and_fix_mapped_measure(self, value):
        value_splitted = value.split(".")
        if len(value_splitted) == 1:
            table_name = self._find_table_name_in_measure_if_not_exists(
                value_splitted[0]
            )
            if table_name is None:
                raise ValueError(
                    "Measure must have table expected format is TableName.measure"
                )
            return f"{table_name}.{value_splitted[0]}"
        return value

    def _validate_and_fix_mapped_dimension(self, value):
        value_splitted = value.split(".")
        if len(value_splitted) == 1:
            table_name = self._find_table_name_in_dimension_if_not_exists(
                value_splitted[0]
            )
            if table_name is None:
                raise ValueError(
                    "Measure must have table expected format is TableName.measure"
                )
            return f"{table_name}.{value_splitted[0]}"
        return value

    def _validate_and_fix_mapped_order(self, value):
        value_splitted = value.split(".")
        if len(value_splitted) == 1:
            table_name = self._find_table_name_in_orders_if_not_exists(
                value_splitted[0]
            )
            if table_name is None:
                raise ValueError(
                    "Measure must have table expected format is TableName.measure"
                )
            return f"{table_name}.{value_splitted[0]}"
        return value

    def _validate_and_fix_mapped_filter(self, value):
        value_splitted = value.split(".")
        if len(value_splitted) == 1:
            table_name = self._find_table_name_in_filter_if_not_exists(
                value_splitted[0]
            )
            if table_name is None:
                raise ValueError(
                    "Measure must have table expected format is TableName.measure"
                )
            return f"{table_name}.{value_splitted[0]}"
        return value

    def _find_table_name_in_filter_if_not_exists(self, filter_name: str):
        """
        Find and add table name if not exists in Measure
        """
        for table in self.schema:
            for dimension in table["dimensions"]:
                if dimension["name"] == filter_name:
                    return table["name"]

        return None

    def _find_table_name_in_measure_if_not_exists(self, measure_name: str):
        """
        Find and add table name if not exists in Measure
        """
        for table in self.schema:
            for measure in table["measures"]:
                if measure["name"] == measure_name:
                    return table["name"]

        return None

    def _find_table_name_in_dimension_if_not_exists(self, dimension_name: str):
        """
        Find and add table name if not exists in Measure
        """
        for table in self.schema:
            for dimension in table["dimensions"]:
                if dimension["name"] == dimension_name:
                    return table["name"]

        return None

    def _find_table_name_in_orders_if_not_exists(self, dimension_name: str):
        """
        Find and add table name if not exists in Measure
        """
        for table in self.schema:
            for dimension in table["dimensions"]:
                if dimension["name"] == dimension_name:
                    return table["name"]

            for measure in table["measures"]:
                if measure["name"] == dimension_name:
                    return table["name"]

        return None

    def _generate_time_dimension_column(self, time_dimension):
        dimension = time_dimension["dimension"]
        granularity = (
            time_dimension["granularity"] if "granularity" in time_dimension else "day"
        )

        if granularity not in self.supported_granularities:
            raise ValueError(
                f"Unsupported granularity '{granularity}'. Supported granularities are: {', '.join(self.supported_granularities)}"
            )

        table = self.find_table(dimension.split(".")[0])["table"]
        dimension_info = self.find_dimension(dimension)
        sql_expr = f"`{table}`.`{dimension_info['sql']}`"

        granularity_sql = {
            "year": f"YEAR({sql_expr})",
            "month": f"DATE_FORMAT({sql_expr}, '%Y-%m')",
            "day": f"DATE_FORMAT({sql_expr}, '%Y-%m-%d')",
            "hour": f"HOUR({sql_expr})",
            "minute": f"MINUTE({sql_expr})",
            "second": f"SECOND({sql_expr})",
        }

        if granularity not in granularity_sql:
            raise ValueError(f"Unhandled granularity: {granularity}")

        return f"{granularity_sql[granularity]} AS {dimension_info['name']}_by_{granularity}"

    def _get_referenced_tables(self, dimensions, time_dimensions, measures, filters):
        return (
            {measure.split(".")[0] for measure in measures}
            | {dim.split(".")[0] for dim in dimensions}
            | {td["dimension"].split(".")[0] for td in time_dimensions}
            | {filter["member"].split(".")[0] for filter in filters}
        )

    def _get_main_table_entry(self, measures, dimensions):
        main_table = (
            measures[0].split(".")[0] if measures else dimensions[0].split(".")[0]
        )
        return next(
            (table for table in self.schema if table["name"] == main_table), None
        )

    def _build_select_clause(self, columns):
        return "SELECT " + ", ".join(columns)

    def _build_from_clause(self, main_table_entry):
        return f" FROM `{main_table_entry['table']}`"

    def _build_joins_clause(self, main_table_entry, referenced_tables):
        sql = ""
        main_table = main_table_entry["name"]

        for table_name in referenced_tables:
            if table_name != main_table:
                table_entry = next(
                    (table for table in self.schema if table["name"] == table_name),
                    None,
                )
                if not table_entry:
                    raise ValueError(f"Table '{table_name}' not found in schema.")
                if "joins" in table_entry and (
                    join := next(
                        (
                            j
                            for j in table_entry["joins"]
                            if j["name"] in {main_table, table_name}
                        ),
                        None,
                    )
                ):
                    join_condition = self.resolve_template_literals(join["sql"])
                    sql += f" {join['join_type'].upper()} JOIN `{table_entry['table']}` ON {join_condition}"

        return sql

    def _build_where_clause(self, filters, time_dimensions):
        filter_statements = [
            self.process_filter(filter)
            for filter in filters
            if self.find_dimension(filter["member"]).get("name") is not None
        ]
        time_dimension_filters = [
            self.resolve_date_range(td) for td in time_dimensions if "dateRange" in td
        ]
        filter_statements.extend(time_dimension_filters)

        return f" WHERE {' AND '.join(filter_statements)}" if filter_statements else ""

    def _build_group_by_clause(self, dimensions, time_dimensions):
        if not (time_dimensions or dimensions):
            return ""

        group_by_dimensions = [
            self.find_dimension(dim)["name"] for dim in dimensions
        ] + [
            f"{self.find_dimension(td['dimension'])['name']}_by_{td.get('granularity', 'day')}"
            for td in time_dimensions
        ]

        return " GROUP BY " + ", ".join(group_by_dimensions)

    def _build_having_clause(self, filters):
        filter_statements = [
            self.process_filter(filter)
            for filter in filters
            if self.find_measure(filter["member"]).get("name") is not None
        ]

        return f" HAVING {' AND '.join(filter_statements)}" if filter_statements else ""

    def _build_order_clause(self, query):
        if "order" not in query or len(query["order"]) == 0:
            return ""

        order_clauses = []
        for order in query["order"]:
            name = None
            if measure := self.find_measure(order["id"]):
                name = measure["name"]

            if (
                name is None
                and "timeDimensions" in query
                and len(query["timeDimensions"]) > 0
            ):
                for time_dimension in query["timeDimensions"]:
                    if (
                        dimension
                        := f"{self.find_dimension(order['id'])['name']}_by_{time_dimension['granularity']}"
                    ):
                        name = dimension

            if name is None and "dimensions" in query and len(query["dimensions"]) > 0:
                if dimension := self.find_dimension(order["id"]):
                    name = dimension["name"]

            if name is None:
                name = (
                    self.find_measure(order["id"]) or self.find_dimension(order["id"])
                )["name"]

            order_clauses.append(f"{name} {order['direction']}")

        return f" ORDER BY {', '.join(order_clauses)}"

    def _build_limit_clause(self, query):
        return f" LIMIT {query['limit']}" if "limit" in query else ""

    def resolve_date_range(self, time_dimension):
        dimension = time_dimension["dimension"]
        date_range = time_dimension["dateRange"]
        table_name = dimension.split(".")[0]
        dimension_info = self.find_dimension(dimension)
        table = self.find_table(table_name)

        if not table or not dimension_info:
            raise ValueError(f"Dimension '{dimension}' not found in schema.")

        table_column = f"`{table['table']}`.`{dimension_info['sql']}`"

        if isinstance(date_range, list) and len(date_range) == 2:
            start_date, end_date = date_range
            return f"{table_column} BETWEEN '{start_date}' AND '{end_date}'"
        else:
            if isinstance(date_range, list) and len(date_range) == 1:
                date_range = date_range[0]

            if date_range not in self.supported_date_ranges:
                raise ValueError(f"Unsupported date range: {date_range}")

            if date_range == "last week":
                return f"{table_column} >= CURRENT_DATE - INTERVAL '1 week' AND {table_column} < CURRENT_DATE"
            elif date_range == "last month":
                return f"{table_column} >= CURRENT_DATE - INTERVAL '1 month' AND {table_column} < CURRENT_DATE"
            elif date_range == "this month":
                return f"{table_column} >= DATE_TRUNC('month', CURRENT_DATE) AND {table_column} < DATE_TRUNC('month', CURRENT_DATE) + INTERVAL '1 month'"
            elif date_range == "this week":
                return f"{table_column} >= DATE_TRUNC('week', CURRENT_DATE) AND {table_column} < DATE_TRUNC('week', CURRENT_DATE) + INTERVAL '1 week'"
            elif date_range == "today":
                return f"{table_column} >= DATE_TRUNC('day', CURRENT_DATE) AND {table_column} < DATE_TRUNC('day', CURRENT_DATE) + INTERVAL '1 day'"
            elif date_range == "this year":
                return f"{table_column} >= DATE_TRUNC('year', CURRENT_DATE) AND {table_column} < DATE_TRUNC('year', CURRENT_DATE) + INTERVAL '1 year'"
            elif date_range == "last year":
                return f"{table_column} >= DATE_TRUNC('year', CURRENT_DATE - INTERVAL '1 year') AND {table_column} < DATE_TRUNC('year', CURRENT_DATE)"

    def process_filter(self, filter):
        required_keys = ["member", "operator", "values"]

        # Check if any required key is missing or if "values" is empty
        if any(key not in filter for key in required_keys) or (
            not filter.get("values")
            and filter.get("operator", None) not in ["set", "notSet"]
        ):
            raise ValueError(f"Invalid filter: {filter}")

        table_name = filter["member"].split(".")[0]
        dimension = self.find_dimension(filter["member"])
        measure = self.find_measure(filter["member"])

        if dimension:
            table_column = f"`{self.find_table(table_name)['table']}`.`{dimension.get('sql', dimension['name'])}`"
        elif measure:
            table_column = f"{measure['type'].upper()}(`{self.find_table(table_name)['table']}`.`{measure.get('sql', measure['name'])}`)"
        else:
            raise ValueError(f"Member '{filter['member']}' not found in schema.")

        operator = filter["operator"]
        values = filter["values"]

        single_value_operators = {
            "equals": "=",
            "notEquals": "!=",
            "contains": "LIKE",
            "notContains": "NOT LIKE",
            "startsWith": "LIKE",
            "endsWith": "LIKE",
            "gt": ">",
            "gte": ">=",
            "lt": "<",
            "lte": "<=",
            "beforeDate": "<",
            "afterDate": ">",
            "in": "IN",
        }

        multi_value_operators = {"equals": "IN", "notEquals": "NOT IN"}

        return self._build_query_condition(
            operator,
            table_column,
            values,
            single_value_operators,
            multi_value_operators,
        )

    def _build_query_condition(
        self,
        operator,
        table_column,
        values,
        single_value_operators,
        multi_value_operators,
    ):
        if operator in single_value_operators:
            if operator in ["equals", "notEquals", "in"]:
                if len(values) == 1:
                    operator_str = "=" if operator == "equals" else "!="
                    return f"{table_column} {operator_str} '{values[0]}'"
                else:
                    operator_str = "IN" if operator in ["equals", "in"] else "NOT IN"
                    formatted_values = "', '".join(values)
                    return f"{table_column} {operator_str} ('{formatted_values}')"

            elif operator in ["contains", "notContains", "startsWith", "endsWith"]:
                pattern = {
                    "contains": f"%{values[0]}%",
                    "notContains": f"%{values[0]}%",
                    "startsWith": f"{values[0]}%",
                    "endsWith": f"%{values[0]}",
                }[operator]
                return f"{table_column} {single_value_operators[operator]} '{pattern}'"

            else:
                value = f"'{values[0]}'" if isinstance(values[0], str) else values[0]
                return f"{table_column} {single_value_operators[operator]} {value}"

        elif operator in multi_value_operators:
            formatted_values = "', '".join(values)
            return f"{table_column} {multi_value_operators[operator]} ('{formatted_values}')"

        elif operator == "set":
            return f"{table_column} IS NOT NULL"

        elif operator == "notSet":
            return f"{table_column} IS NULL"

        elif operator in ["inDateRange", "notInDateRange"]:
            if len(values) != 2:
                raise ValueError(f"Invalid number of values for '{operator}' operator.")
            range_operator = "BETWEEN" if operator == "inDateRange" else "NOT BETWEEN"
            return f"{table_column} {range_operator} '{values[0]}' AND '{values[1]}'"

        else:
            raise ValueError(f"Unsupported operator: {operator}")

    def resolve_template_literals(self, template):
        def replace_column(match):
            table, column = match.group(1).split(".")
            new_table = self.find_table(table)
            if not new_table:
                raise ValueError(f"Table '{table}' not found in schema.")
            new_column = next(
                (dim for dim in new_table["dimensions"] if dim["name"] == column), None
            )
            if not new_column:
                raise ValueError(f"Column '{column}' not found in schema.")
            return f"`{new_table['table']}`.`{new_column['sql']}`"

        return re.sub(r"\$\{([^}]+)\}", replace_column, template)

    def find_table(self, table_name):
        return next((table for table in self.schema if table["name"] == table_name), {})

    def find_dimension(self, dimension):
        table_name, dim_name = dimension.split(".")
        table = self.find_table(table_name)
        dim = next(
            (dim for dim in table.get("dimensions", []) if dim.get("name") == dim_name),
            {},
        )
        return dim

    def find_measure(self, measure):
        table_name, measure_name = measure.split(".")
        table = self.find_table(table_name)
        meas = next(
            (
                meas
                for meas in table.get("measures", [])
                if meas.get("name") == measure_name
            ),
            {},
        )
        return meas



import logging
from typing import Optional, Union

from .. import Logger
from ._output_types import (
    DataFrameOutputType,
    DefaultOutputType,
    NumberOutputType,
    PlotOutputType,
    StringOutputType,
)

output_types_map = {
    "number": NumberOutputType,
    "dataframe": DataFrameOutputType,
    "plot": PlotOutputType,
    "string": StringOutputType,
}


def output_type_factory(
    output_type: str = None, logger: Optional[Logger] = None
) -> Union[
    NumberOutputType,
    DataFrameOutputType,
    PlotOutputType,
    StringOutputType,
    DefaultOutputType,
]:
    """
    Factory function to get appropriate instance for output type.

    Uses `output_types_map` to determine the output type class.

    Args:
        output_type (Optional[str]): A name of the output type.
            Defaults to None, an instance of `DefaultOutputType` will be
            returned.
        logger (Optional[str]): If passed, collects logs about correctness
            of the `output_type` argument and what kind of OutputType
            is created.

    Returns:
        (Union[
            NumberOutputType,
            DataFrameOutputType,
            PlotOutputType,
            StringOutputType,
            DefaultOutputType
        ]): An instance of the output type.
    """
    if output_type is not None and output_type not in output_types_map and logger:
        possible_types_msg = ", ".join(f"'{type_}'" for type_ in output_types_map)
        logger.log(
            f"Unknown value for the parameter `output_type`: '{output_type}'."
            f"Possible values are: {possible_types_msg} and None for default "
            f"output type (miscellaneous).",
            level=logging.WARNING,
        )

    output_type_helper = output_types_map.get(output_type, DefaultOutputType)()

    if logger:
        logger.log(
            f"{output_type_helper.__class__} is going to be used.", level=logging.DEBUG
        )

    return output_type_helper



"""
Constants used in the pandasai package.

It includes Start & End Code tags, Whitelisted Python Packages and
While List Builtin Methods.

"""

# Default directory to store chart if user doesn't provide any
DEFAULT_CHART_DIRECTORY = "exports/charts"

# Default directory for cache
DEFAULT_CACHE_DIRECTORY = "cache"

# Default permissions for files and directories
DEFAULT_FILE_PERMISSIONS = 0o755

# Token needed to invalidate the cache after breaking changes
CACHE_TOKEN = "pandasai1"

# List of Python builtin libraries that are added to the environment by default.
WHITELISTED_BUILTINS = [
    "abs",
    "all",
    "any",
    "ascii",
    "bin",
    "bool",
    "bytearray",
    "bytes",
    "callable",
    "chr",
    "classmethod",
    "complex",
    "delattr",
    "dict",
    "dir",
    "divmod",
    "enumerate",
    "filter",
    "float",
    "format",
    "frozenset",
    "getattr",
    "hasattr",
    "hash",
    "help",
    "hex",
    "id",
    "int",
    "isinstance",
    "issubclass",
    "iter",
    "len",
    "list",
    "locals",
    "map",
    "max",
    "memoryview",
    "min",
    "next",
    "object",
    "oct",
    "ord",
    "pow",
    "print",
    "property",
    "range",
    "repr",
    "reversed",
    "round",
    "set",
    "setattr",
    "slice",
    "sorted",
    "staticmethod",
    "str",
    "sum",
    "super",
    "tuple",
    "type",
    "vars",
    "zip",
]

# List of Python packages that are whitelisted for import in generated code
WHITELISTED_LIBRARIES = [
    "sklearn",
    "statsmodels",
    "seaborn",
    "plotly",
    "ggplot",
    "matplotlib",
    "numpy",
    "datetime",
    "json",
    "io",
    "base64",
    "scipy",
    "streamlit",
    "modin",
    "scikit-learn",
]

PANDASBI_SETUP_MESSAGE = (
    "The api_key client option must be set either by passing api_key to the client "
    "or by setting the PANDASAI_API_KEY environment variable. To get the key follow below steps:\n"
    "1. Go to https://www.pandabi.ai and sign up\n"
    "2. From settings go to API keys and copy\n"
    "3. Set environment variable like os.environ['PANDASAI_API_KEY'] = '$2a$10$flb7....'"
)



import sys

from dotenv import load_dotenv as _load_dotenv

from .path import find_closest


def load_dotenv():
    """
    Load the .env file from the root folder of the project
    """
    try:
        dotenv_path = find_closest(".env")
        _load_dotenv(dotenv_path=dotenv_path)
    except ValueError:
        pass


def is_running_in_console() -> bool:
    """
    Check if the code is running in console or not.

    Returns:
        bool: True if running in console else False
    """

    return sys.stdout.isatty()



"""
A smart dataframe class is a wrapper around the pandas/polars dataframe that allows you
to query it using natural language. It uses the LLMs to generate Python code from
natural language and then executes it on the dataframe.

Example:
    ```python
    from pandasai.smart_dataframe import SmartDataframe
    from pandasai.llm.openai import OpenAI

    df = pd.read_csv("examples/data/Loan payments data.csv")
    llm = OpenAI()

    df = SmartDataframe(df, config={"llm": llm})
    response = df.chat("What is the average loan amount?")
    print(response)
    # The average loan amount is $15,000.
    ```
"""

import uuid
from typing import Any, List, Optional, Union

from pandasai.agent import Agent
from pandasai.skills import Skill

from ..helpers.cache import Cache
from ..helpers.df_info import DataFrameType
from ..schemas.df_config import Config


class SmartDatalake:
    def __init__(
        self,
        dfs: List[Union[DataFrameType, Any]],
        config: Optional[Union[Config, dict]] = None,
    ):
        """
        Args:
            dfs (List[Union[DataFrameType, Any]]): List of dataframes to be used
            config (Union[Config, dict], optional): Config to be used. Defaults to None.
        """
        self._agent = Agent(dfs, config=config)

    def add_skills(self, *skills: Skill):
        """
        Add Skills to PandasAI
        """
        self._agent.add_skills(*skills)

    def chat(self, query: str, output_type: Optional[str] = None):
        """
        Run a query on the dataframe.

        Args:
            query (str): Query to run on the dataframe
            output_type (Optional[str]): Add a hint for LLM which
                type should be returned by `analyze_data()` in generated
                code. Possible values: "number", "dataframe", "plot", "string":
                    * number - specifies that user expects to get a number
                        as a response object
                    * dataframe - specifies that user expects to get
                        pandas/polars dataframe as a response object
                    * plot - specifies that user expects LLM to build
                        a plot
                    * string - specifies that user expects to get text
                        as a response object
                If none `output_type` is specified, the type can be any
                of the above or "text".

        Raises:
            ValueError: If the query is empty
        """
        return self._agent.chat(query, output_type)

    def clear_memory(self):
        """
        Clears the memory
        """
        self._agent.clear_memory()

    @property
    def last_prompt(self):
        return self._agent.last_prompt

    @property
    def last_prompt_id(self) -> uuid.UUID:
        """Return the id of the last prompt that was run."""
        if self._agent.last_prompt_id is None:
            raise ValueError("Pandas AI has not been run yet.")
        return self._agent.last_prompt_id

    @property
    def logs(self):
        return self._agent.logger.logs

    @property
    def logger(self):
        return self._agent.logger

    @logger.setter
    def logger(self, logger):
        self._agent.logger = logger

    @property
    def config(self):
        return self._agent.context.config

    @property
    def cache(self):
        return self._agent.context.cache

    @property
    def verbose(self):
        return self._agent.context.config.verbose

    @verbose.setter
    def verbose(self, verbose: bool):
        self._agent.context.config.verbose = verbose
        self._agent.logger.verbose = verbose

    @property
    def save_logs(self):
        return self._agent.context.config.save_logs

    @save_logs.setter
    def save_logs(self, save_logs: bool):
        self._agent.context.config.save_logs = save_logs
        self._agent.logger.save_logs = save_logs

    @property
    def enforce_privacy(self):
        return self._agent.context.config.enforce_privacy

    @enforce_privacy.setter
    def enforce_privacy(self, enforce_privacy: bool):
        self._agent.context.config.enforce_privacy = enforce_privacy

    @property
    def enable_cache(self):
        return self._agent.context.config.enable_cache

    @enable_cache.setter
    def enable_cache(self, enable_cache: bool):
        self._agent.context.config.enable_cache = enable_cache
        if enable_cache:
            if self.cache is None:
                self._cache = Cache()
        else:
            self._cache = None

    @property
    def use_error_correction_framework(self):
        return self._agent.context.config.use_error_correction_framework

    @use_error_correction_framework.setter
    def use_error_correction_framework(self, use_error_correction_framework: bool):
        self._agent.context.config.use_error_correction_framework = (
            use_error_correction_framework
        )

    @property
    def custom_prompts(self):
        return self._agent.context.config.custom_prompts

    @custom_prompts.setter
    def custom_prompts(self, custom_prompts: dict):
        self._agent.context.config.custom_prompts = custom_prompts

    @property
    def save_charts(self):
        return self._agent.context.config.save_charts

    @save_charts.setter
    def save_charts(self, save_charts: bool):
        self._agent.context.config.save_charts = save_charts

    @property
    def save_charts_path(self):
        return self._agent.context.config.save_charts_path

    @save_charts_path.setter
    def save_charts_path(self, save_charts_path: str):
        self._agent.context.config.save_charts_path = save_charts_path

    @property
    def last_code_generated(self):
        return self._agent.last_code_generated

    @property
    def last_code_executed(self):
        return self._agent.last_code_executed

    @property
    def last_result(self):
        return self._agent.last_result

    @property
    def last_error(self):
        return self._agent.last_error

    @property
    def dfs(self):
        return self._agent.context.dfs

    @property
    def memory(self):
        return self._agent.context.memory

    @property
    def last_query_log_id(self):
        return self._agent.last_query_log_id



"""Google Gemini LLM

This module is to run the Google Gemini API hosted and maintained by Google.
To read more on Google Gemini follow
https://ai.google.dev/docs/gemini_api_overview.

Example:
    Use below example to call GoogleGemini Model

    >>> from pandasai.llm.google_gemini import GoogleGemini
"""
from typing import Any, Optional

from ..exceptions import APIKeyNotFoundError
from ..helpers.memory import Memory
from ..helpers.optional import import_dependency
from .base import BaseGoogle


class GoogleGemini(BaseGoogle):
    """Google Gemini LLM
    BaseGoogle class is extended for Google Gemini model. The default and only model
    support at the moment is models/gemini-pro.

    """

    model: str = "models/gemini-pro"
    google_gemini: Any

    def __init__(self, api_key: str, **kwargs):
        """
        __init__ method of GoogleGemini Class
        Args:
            api_key (str): API Key
            **kwargs: Extended Parameters inferred from BaseGoogle Class
        """
        self._configure(api_key=api_key)
        self._set_params(**kwargs)

    def _configure(self, api_key: str):
        """
        Configure Google Gemini API Key
        Args:
            api_key (str): A string of API keys generated from Google Cloud.

        Returns:
            None.
        """

        if not api_key:
            raise APIKeyNotFoundError("Google Gemini API key is required")

        err_msg = "Install google-generativeai >= 0.3 for Google Gemini API"
        self.google_gemini = import_dependency("google.generativeai", extra=err_msg)

        self.google_gemini.configure(api_key=api_key)
        final_model = self.google_gemini.GenerativeModel(self.model)
        self.google_gemini = final_model

    def _valid_params(self):
        """Returns if the Parameters are valid or Not"""
        return super()._valid_params() + ["model"]

    def _validate(self):
        """
        A method to Validate the Model

        """

        super()._validate()

        if not self.model:
            raise ValueError("model is required.")

    def _generate_text(self, prompt: str, memory: Optional[Memory] = None) -> str:
        """
        Generates text for prompt.

        Args:
            prompt (str): A string representation of the prompt.

        Returns:
            str: LLM response.

        """
        self._validate()
        updated_prompt = self.prepend_system_prompt(prompt, memory)

        self.last_prompt = updated_prompt
        completion = self.google_gemini.generate_content(
            contents=prompt,
            generation_config=dict(
                {
                    "temperature": self.temperature,
                    "top_p": self.top_p,
                    "top_k": self.top_k,
                    "max_output_tokens": self.max_output_tokens,
                }
            ),
        )

        return completion.text

    @property
    def type(self) -> str:
        return "google-gemini"



"""
SnowFlake connectors are used to connect to SnowFlake Data Cloud.
"""

from functools import cache
from typing import Union

from sqlalchemy import create_engine

import pandasai.pandas as pd

from ...connectors.base import BaseConnectorConfig
from ...connectors.sql import SQLBaseConnectorConfig, SQLConnector


class SnowFlakeConnectorConfig(SQLBaseConnectorConfig):
    """
    Connector configuration for SnowFlake.
    """

    account: str
    database: str
    username: str
    password: str
    dbSchema: str
    warehouse: str


class SnowFlakeConnector(SQLConnector):
    """
    SnowFlake connectors are used to connect to SnowFlake Data Cloud.
    """

    def __init__(
        self,
        config: Union[SnowFlakeConnectorConfig, dict],
        **kwargs,
    ):
        """
        Initialize the SnowFlake connector with the given configuration.

        Args:
            config (ConnectorConfig): The configuration for the SnowFlake connector.
        """
        config["dialect"] = "snowflake"

        if isinstance(config, dict):
            snowflake_env_vars = {
                "account": "SNOWFLAKE_HOST",
                "database": "SNOWFLAKE_DATABASE",
                "warehouse": "SNOWFLAKE_WAREHOUSE",
                "dbSchema": "SNOWFLAKE_SCHEMA",
                "username": "SNOWFLAKE_USERNAME",
                "password": "SNOWFLAKE_PASSWORD",
            }
            config = self._populate_config_from_env(config, snowflake_env_vars)

        super().__init__(config, **kwargs)

    def _load_connector_config(self, config: Union[BaseConnectorConfig, dict]):
        return SnowFlakeConnectorConfig(**config)

    def _init_connection(self, config: SnowFlakeConnectorConfig):
        """
        Initialize Database Connection

        Args:
            config (SQLConnectorConfig): Configurations to load database

        """
        self._engine = create_engine(
            f"{config.dialect}://{config.username}:{config.password}@{config.account}/?warehouse={config.warehouse}&database={config.database}&schema={config.dbSchema}"
        )

        self._connection = self._engine.connect()

    @cache
    def head(self, n: int = 5) -> pd.DataFrame:
        """
        Return the head of the data source that the connector is connected to.
        This information is passed to the LLM to provide the schema of the data source.

        Returns:
            DataFrame: The head of the data source.
        """

        if self.logger:
            self.logger.log(
                f"Getting head of {self.config.table} "
                f"using dialect {self.config.dialect}"
            )

        # Run a SQL query to get all the columns names and 5 random rows
        query = self._build_query(limit=n, order="RANDOM()")

        # Return the head of the data source
        return pd.read_sql(query, self._connection)

    def __repr__(self):
        """
        Return the string representation of the SnowFlake connector.

        Returns:
            str: The string representation of the SnowFlake connector.
        """
        return (
            f"<{self.__class__.__name__} dialect={self.config.dialect} "
            f"Account={self.config.account} "
            f"warehouse={self.config.warehouse} "
            f"database={self.config.database} schema={str(self.config.dbSchema)}  "
            f"table={self.config.table}>"
        )

    def equals(self, other):
        if isinstance(other, self.__class__):
            return (
                self.config.dialect,
                self.config.dbSchema,
                self.config.warehouse,
                self.config.account,
            ) == (
                other.config.dialect,
                other.config.dbSchema,
                other.config.warehouse,
                other.config.account,
            )
        return False



import threading
from importlib import reload

_engine = "pandas"
_lock: threading.RLock = threading.RLock()


def set_pd_engine(engine: str = "pandas"):
    global _engine
    if engine.lower() not in ("modin", "pandas"):
        raise ValueError(
            f"Unknown engine {engine}. Valid options are ('modin', 'pandas')"
        )

    if engine != _engine:
        with _lock:
            _engine = engine
            _reload_pd()


def _reload_pd():
    import pandasai

    reload(pandasai.pandas)



from .base import BasePrompt


class CorrectErrorPrompt(BasePrompt):
    """Prompt to generate Python code from a dataframe."""

    template_path = "correct_error_prompt.tmpl"

    def to_json(self):
        context = self.props["context"]
        code = self.props["code"]
        error = self.props["error"]
        memory = context.memory
        conversations = memory.to_json()

        system_prompt = memory.get_system_prompt()

        # prepare datasets
        datasets = [dataset.to_json() for dataset in context.dfs]

        return {
            "datasets": datasets,
            "conversation": conversations,
            "system_prompt": system_prompt,
            "error": {
                "code": code,
                "error_trace": str(error),
                "exception_type": "Exception",
            },
            "config": {"direct_sql": context.config.direct_sql},
        }



"""
Logger class

This class is used to log messages to the console and/or a file.

Example:
    ```python
    from pandasai.helpers.logger import Logger

    logger = Logger()
    logger.log("Hello, world!")
    # 2021-08-01 12:00:00 [INFO] Hello, world!

    logger.logs
    #["Hello, world!"]
    ```
"""

import inspect
import logging
import sys
import time
from typing import List

from pandasai.helpers.telemetry import scarf_analytics
from pandasai.pydantic import BaseModel

from .path import find_closest


class Log(BaseModel):
    """Log class"""

    msg: str
    level: int


class Logger:
    """Logger class"""

    _logs: List[Log]
    _logger: logging.Logger
    _verbose: bool
    _last_time: float

    def __init__(self, save_logs: bool = True, verbose: bool = False):
        """Initialize the logger"""
        self._logs = []
        self._verbose = verbose
        self._last_time = time.time()

        if save_logs:
            try:
                filename = find_closest("pandasai.log")
            except ValueError:
                filename = "pandasai.log"
            handlers = [logging.FileHandler(filename)]
        else:
            handlers = []

        if verbose:
            handlers.append(logging.StreamHandler(sys.stdout))

        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s [%(levelname)s] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            handlers=handlers,
        )
        self._logger = logging.getLogger(__name__)

    def log(self, message: str, level: int = logging.INFO):
        """Log a message"""

        if level == logging.INFO:
            self._logger.info(message)
        elif level == logging.WARNING:
            self._logger.warning(message)
        elif level == logging.ERROR:
            self._logger.error(message)
        elif level == logging.CRITICAL:
            self._logger.critical(message)

        self._logs.append(
            {
                "msg": message,
                "level": logging.getLevelName(level),
                "time": self._calculate_time_diff(),
                "source": self._invoked_from(),
            }
        )

    def _invoked_from(self, level: int = 5) -> str:
        """Return the name of the class that invoked the logger"""
        calling_class = None
        for frame_info in inspect.stack()[1:]:
            frame_locals = frame_info[0].f_locals
            calling_instance = frame_locals.get("self")
            if calling_instance and calling_instance.__class__ != self.__class__:
                calling_class = calling_instance.__class__.__name__
                break
            level -= 1
            if level <= 0:
                break
        return calling_class

    def _calculate_time_diff(self):
        """Calculate the time difference since the last log"""
        time_diff = time.time() - self._last_time
        self._last_time = time.time()
        return time_diff

    @property
    def logs(self) -> List[str]:
        """Return the logs"""
        return self._logs

    @property
    def verbose(self) -> bool:
        """Return the verbose flag"""
        return self._verbose

    @verbose.setter
    def verbose(self, verbose: bool):
        """Set the verbose flag"""
        self._verbose = verbose
        self._logger.handlers = []
        if verbose:
            self._logger.addHandler(logging.StreamHandler(sys.stdout))
        else:
            # remove the StreamHandler if it exists
            for handler in self._logger.handlers:
                if isinstance(handler, logging.StreamHandler):
                    self._logger.removeHandler(handler)

    @property
    def save_logs(self) -> bool:
        """Return the save_logs flag"""
        return len(self._logger.handlers) > 0

    @save_logs.setter
    def save_logs(self, save_logs: bool):
        """Set the save_logs flag"""
        if save_logs and not self.save_logs:
            filename = find_closest("pandasai.log")
            self._logger.addHandler(logging.FileHandler(filename))
        elif not save_logs and self.save_logs:
            # remove the FileHandler if it exists
            for handler in self._logger.handlers:
                if isinstance(handler, logging.FileHandler):
                    self._logger.removeHandler(handler)


scarf_analytics()



from typing import Any, List, Optional, TypedDict

from pandasai.constants import DEFAULT_CHART_DIRECTORY
from pandasai.helpers.dataframe_serializer import DataframeSerializerType
from pandasai.pydantic import BaseModel, Field, validator

from ..llm import LLM, BambooLLM, LangchainLLM


class LogServerConfig(TypedDict):
    server_url: str
    api_key: str


class Config(BaseModel):
    save_logs: bool = True
    verbose: bool = False
    enforce_privacy: bool = False
    enable_cache: bool = True
    use_error_correction_framework: bool = True
    open_charts: bool = True
    save_charts: bool = False
    save_charts_path: str = DEFAULT_CHART_DIRECTORY
    custom_whitelisted_dependencies: List[str] = Field(default_factory=list)
    max_retries: int = 3
    lazy_load_connector: bool = True
    response_parser: Any = None
    llm: LLM = None
    data_viz_library: Optional[str] = ""
    log_server: LogServerConfig = None
    direct_sql: bool = False
    dataframe_serializer: DataframeSerializerType = DataframeSerializerType.CSV

    class Config:
        arbitrary_types_allowed = True

    @validator("llm", always=True)
    def validate_llm(cls, llm) -> LLM:
        if not isinstance(llm, (LLM, LangchainLLM)):  # also covers llm is None
            return BambooLLM()
        return llm



from abc import abstractmethod


class AbstractRelation:
    @abstractmethod
    def to_string(self):
        raise NotImplementedError


class PrimaryKey(AbstractRelation):
    def __init__(self, name):
        self.name = name

    def to_string(self):
        return f"PRIMARY KEY ({self.name})"


class ForeignKey(AbstractRelation):
    def __init__(self, field, foreign_table, foreign_table_field):
        self.field = field
        self.foreign_table_field = foreign_table_field
        self.foreign_table = foreign_table

    def to_string(self):
        return f"FOREIGN KEY ({self.field}) REFERENCES {self.foreign_table}({self.foreign_table_field})"



"""OpenAI LLM API

This module is to run the OpenAI API using OpenAI API.

Example:
    Use below example to call OpenAI Model

    >>> from pandasai.llm.openai import OpenAI
"""

import os
from typing import Any, Dict, Optional

import openai

from ..exceptions import APIKeyNotFoundError, UnsupportedModelError
from ..helpers import load_dotenv
from ..helpers.openai import is_openai_v1
from .base import BaseOpenAI

load_dotenv()


class OpenAI(BaseOpenAI):
    """OpenAI LLM using BaseOpenAI Class.

    An API call to OpenAI API is sent and response is recorded and returned.
    The default chat model is **gpt-3.5-turbo**.
    The list of supported Chat models includes ["gpt-4", "gpt-4-0613", "gpt-4-32k",
     "gpt-4-32k-0613", "gpt-3.5-turbo", "gpt-3.5-turbo-16k", "gpt-3.5-turbo-0613",
     "gpt-3.5-turbo-16k-0613", "gpt-3.5-turbo-instruct"].
    The list of supported Completion models includes "gpt-3.5-turbo-instruct" and
     "text-davinci-003" (soon to be deprecated).
    """

    _supported_chat_models = [
        "gpt-3.5-turbo",
        "gpt-3.5-turbo-0125",
        "gpt-3.5-turbo-1106",
        "gpt-3.5-turbo-0613",
        "gpt-3.5-turbo-16k",
        "gpt-3.5-turbo-16k-0613",
        "gpt-4",
        "gpt-4-0125-preview",
        "gpt-4-1106-preview",
        "gpt-4-0613",
        "gpt-4-32k",
        "gpt-4-32k-0613",
        "gpt-4-turbo-preview",
        "gpt-4o",
        "gpt-4o-2024-05-13",
        "gpt-4o-mini",
        "gpt-4o-mini-2024-07-18",
    ]
    _supported_completion_models = ["gpt-3.5-turbo-instruct"]

    model: str = "gpt-4o-mini"

    def __init__(
        self,
        api_token: Optional[str] = None,
        **kwargs,
    ):
        """
        __init__ method of OpenAI Class

        Args:
            api_token (str): API Token for OpenAI platform.
            **kwargs: Extended Parameters inferred from BaseOpenAI class

        """
        self.api_token = api_token or os.getenv("OPENAI_API_KEY") or None

        if not self.api_token:
            raise APIKeyNotFoundError("OpenAI API key is required")

        self.api_base = (
            kwargs.get("api_base") or os.getenv("OPENAI_API_BASE") or self.api_base
        )
        self.openai_proxy = kwargs.get("openai_proxy") or os.getenv("OPENAI_PROXY")
        if self.openai_proxy:
            openai.proxy = {"http": self.openai_proxy, "https": self.openai_proxy}

        self._set_params(**kwargs)
        # set the client
        model_name = self.model.split(":")[1] if "ft:" in self.model else self.model
        if model_name in self._supported_chat_models:
            self._is_chat_model = True
            self.client = (
                openai.OpenAI(**self._client_params).chat.completions
                if is_openai_v1()
                else openai.ChatCompletion
            )
        elif model_name in self._supported_completion_models:
            self._is_chat_model = False
            self.client = (
                openai.OpenAI(**self._client_params).completions
                if is_openai_v1()
                else openai.Completion
            )
        else:
            raise UnsupportedModelError(self.model)

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling OpenAI API"""
        return {
            **super()._default_params,
            "model": self.model,
        }

    @property
    def type(self) -> str:
        return "openai"



from dataclasses import dataclass


@dataclass
class JudgePipelineInput:
    query: str
    code: str

    def __init__(self, query: str, code: str) -> None:
        self.query = query
        self.code = code



from typing import Union

import pandas as pd


def _import_modin():
    try:
        import modin.pandas as pd
    except ImportError as e:
        raise ImportError(
            "Could not import modin, please install with " "`pip install modin`."
        ) from e
    return pd


def _import_polars():
    try:
        import polars as pl
    except ImportError as e:
        raise ImportError(
            "Could not import polars, please install with " "`pip install polars`."
        ) from e
    return pl


DataFrameType = Union[pd.DataFrame, str]

polars_imported = False
modin_imported = False
try:
    pl = _import_polars()

    polars_imported = True
    DataFrameType = Union[DataFrameType, pl.DataFrame]
except ImportError:
    pass

try:
    mpd = _import_modin()

    modin_imported = True
    DataFrameType = Union[DataFrameType, mpd.DataFrame]
except ImportError:
    pass


def df_type(df: DataFrameType) -> Union[str, None]:
    """
    Returns the type of the dataframe.

    Args:
        df (DataFrameType): Pandas, Modin or Polars dataframe

    Returns:
        str: Type of the dataframe
    """
    if polars_imported and isinstance(df, pl.DataFrame):
        return "polars"
    elif modin_imported and isinstance(df, mpd.DataFrame):
        return "modin"
    elif isinstance(df, pd.DataFrame):
        return "pandas"
    else:
        return None



import ast


class AssignmentVisitor(ast.NodeVisitor):
    def __init__(self):
        self.assignment_nodes = []

    def visit_Assign(self, node):  # noqa: N802
        self.assignment_nodes.append(node)
        self.generic_visit(node)


class CallVisitor(ast.NodeVisitor):
    def __init__(self):
        self.call_nodes = []

    def visit_Call(self, node):  # noqa: N802
        self.call_nodes.append(node)
        self.generic_visit(node)



import uuid
from dataclasses import dataclass


@dataclass
class ChatPipelineInput:
    """
    Contain all the data needed by the chat pipeline
    """

    query: str
    output_type: str
    instance: str
    conversation_id: uuid.UUID
    prompt_id: uuid.UUID

    def __init__(
        self,
        query: str,
        output_type: str,
        conversation_id: uuid.UUID,
        prompt_id: uuid.UUID,
    ) -> None:
        self.query = query
        self.output_type = output_type
        self.conversation_id = conversation_id
        self.prompt_id = prompt_id



from typing import Iterable, List, Optional, Union

from pandasai.helpers.logger import Logger
from pandasai.helpers.request import Session
from pandasai.vectorstores.vectorstore import VectorStore


class BambooVectorStore(VectorStore):
    """
    Implementation of ChromeDB vector store
    """

    _logger: Logger

    def __init__(
        self,
        endpoint_url: Optional[str] = None,
        api_key: Optional[str] = None,
        logger: Optional[Logger] = None,
        max_samples: int = 1,
    ) -> None:
        self._max_samples = max_samples
        self._logger = logger or Logger()
        self._session = Session(endpoint_url, api_key, logger)

    def add_question_answer(self, queries: Iterable[str], codes: Iterable[str]) -> bool:
        """
        Add question and answer(code) to the training set
        Args:
            queries: string of question
            codes: str
        """
        self._session.post("/training-data", json={"query": queries, "code": codes})
        return True

    def add_docs(self, docs: Iterable[str]) -> bool:
        """
        Add docs to the training set
        Args:
            docs: Iterable of strings to add to the vectorstore.
            ids: Optional Iterable of ids associated with the texts.
            metadatas: Optional list of metadatas associated with the texts.
            kwargs: vectorstore specific parameters

        Returns:
            List of ids from adding the texts into the vectorstore.
        """
        self._session.post("/training-docs", json={"docs": docs})
        return True

    def get_relevant_qa_documents(self, question: str, k: int = None) -> List[dict]:
        """
        Returns relevant question answers based on search
        """
        k = k or self._max_samples

        try:
            docs = self._session.get(
                "/training-data/qa/relevant-qa", params={"query": question, "count": k}
            )
            return docs["docs"]
        except Exception:
            self._logger.log("Querying without using training data.")
            return []

    def get_relevant_docs_documents(
        self, question: str, k: Union[int, None] = 3
    ) -> List[str]:
        """
        Returns relevant question answers documents only
        Args:
            question (_type_): list of documents
        """
        k = k or self._max_samples
        try:
            docs = self._session.get(
                "/training-docs/docs/relevant-docs",
                params={"query": question, "count": k},
            )
            return docs["docs"]
        except Exception:
            self._logger.log("Querying without using training docs.")
            return []



"""Helper functions to save charts to a file, if plt.show() is called."""
from pathlib import Path

from .logger import Logger


def add_save_chart(
    code: str,
    logger: Logger,
    file_name: str,
    save_charts_path_str: str = None,
) -> str:
    """
    Add line to code that save charts to a file, if plt.show() is called.

    Args:
        code (str): Code to add line to.
        logger (logging.Logger): Logger to use.
        file_name (str): Name of folder to save charts to.
        save_charts_path (str): User Defined Path to save Charts

    Returns:
        str: Code with line added.

    """
    save_charts_path = Path(save_charts_path_str) if save_charts_path_str else None

    if save_charts_path is not None:
        save_charts_path.mkdir(parents=True, exist_ok=True)

        save_charts_file = save_charts_path / f"{file_name}.png"

        if "temp_chart.png" in code:
            code = code.replace("temp_chart.png", save_charts_file.as_posix())
            logger.log(f"Saving charts to {save_charts_file}")

    return code



"""
Pandas connector class to handle csv, parquet, xlsx files and pandas dataframes.
"""

import hashlib
from functools import cache, cached_property
from typing import Union

import duckdb
import sqlglot
from pydantic import BaseModel

import pandasai.pandas as pd
from pandasai.exceptions import PandasConnectorTableNotFound

from ..helpers.data_sampler import DataSampler
from ..helpers.file_importer import FileImporter
from ..helpers.logger import Logger
from .base import BaseConnector


class PandasConnectorConfig(BaseModel):
    """
    Pandas Connector configuration.
    """

    original_df: Union[pd.DataFrame, pd.Series, str, list, dict]

    class Config:
        arbitrary_types_allowed = True


class PandasConnector(BaseConnector):
    """
    Pandas connector class to handle csv, parquet, xlsx files and pandas dataframes.
    """

    pandas_df = pd.DataFrame
    _logger: Logger = None
    _additional_filters: list[list[str]] = None

    def __init__(
        self,
        config: Union[PandasConnectorConfig, dict],
        **kwargs,
    ):
        """
        Initialize the Pandas connector with the given configuration.

        Args:
            config (PandasConnectorConfig): The configuration for the Pandas connector.
        """
        super().__init__(config, **kwargs)

        self._load_df(self.config.original_df)
        self.sql_enabled = False

    def _load_df(self, df: Union[pd.DataFrame, pd.Series, str, list, dict]):
        """
        Load the dataframe from a file or pandas dataframe.

        Args:
            df (Union[pd.DataFrame, pd.Series, str, list, dict]): The dataframe to load.
        """
        if isinstance(df, pd.Series):
            self.pandas_df = df.to_frame()
        elif isinstance(df, pd.DataFrame):
            self.pandas_df = df
        elif isinstance(df, (list, dict)):
            try:
                self.pandas_df = pd.DataFrame(df)
            except Exception as e:
                raise ValueError(
                    "Invalid input data. We cannot convert it to a dataframe."
                ) from e
        elif isinstance(df, str):
            self.pandas_df = FileImporter.import_from_file(df)
        else:
            raise ValueError("Invalid input data. We cannot convert it to a dataframe.")

    def _load_connector_config(
        self, config: Union[PandasConnectorConfig, dict]
    ) -> PandasConnectorConfig:
        """
        Loads passed Configuration to object

        Args:
            config (PandasConnectorConfig): Construct config in structure

            Returns:
                config: PandasConnectorConfig
        """
        return PandasConnectorConfig(**config)

    @cache
    def head(self, n: int = 5) -> pd.DataFrame:
        """
        Return the head of the data source that the connector is connected to.
        This information is passed to the LLM to provide the schema of the
        data source.
        """
        sampler = DataSampler(self.pandas_df)
        return sampler.sample(n)

    @cache
    def execute(self) -> pd.DataFrame:
        """
        Execute the given query on the data source that the connector is
        connected to.
        """
        return self.pandas_df

    @cached_property
    def rows_count(self):
        """
        Return the number of rows in the data source that the connector is
        connected to.
        """
        return len(self.pandas_df)

    @cached_property
    def columns_count(self):
        """
        Return the number of columns in the data source that the connector is
        connected to.
        """
        return len(self.pandas_df.columns)

    @property
    def column_hash(self):
        """
        Return the hash code that is unique to the columns of the data source
        that the connector is connected to.
        """
        columns_str = "".join(self.pandas_df.columns)
        hash_object = hashlib.sha256(columns_str.encode())
        return hash_object.hexdigest()

    @cached_property
    def path(self):
        """
        Return the path of the data source that the connector is connected to.
        """
        pass

    @property
    def fallback_name(self):
        """
        Return the name of the table that the connector is connected to.
        """
        pass

    @property
    def type(self):
        return "pd.DataFrame"

    def equals(self, other: BaseConnector):
        """
        Return whether the data source that the connector is connected to is
        equal to the other data source.
        """
        return self._original_df.equals(other._original_df)

    def enable_sql_query(self, table_name=None):
        if not table_name and not self.name:
            raise PandasConnectorTableNotFound("Table name not found!")

        table = table_name or self.name

        # Check if the table already exists in DuckDB
        existing_tables = duckdb.query("SHOW TABLES").fetchall()

        # If the table already exists, drop it
        if table in [t[0] for t in existing_tables]:
            duckdb.query(f"DROP TABLE {table}")

        duckdb_relation = duckdb.from_df(self.pandas_df)
        duckdb_relation.create(table)
        self.sql_enabled = True
        self.name = table

    def execute_direct_sql_query(self, sql_query):
        if not self.sql_enabled:
            self.enable_sql_query()

        sql_query = sqlglot.transpile(sql_query, read="mysql", write="duckdb")[0]
        return duckdb.query(sql_query).df()

    @property
    def cs_table_name(self):
        return self.name



""" Prompt to explain code generation by the LLM"""
import pandasai.pandas as pd
from pandasai.helpers.dataframe_serializer import (
    DataframeSerializer,
    DataframeSerializerType,
)

from .generate_python_code import (
    CurrentCodePrompt,
    GeneratePythonCodePrompt,
    SimpleReasoningPrompt,
)


class DirectSQLPrompt(GeneratePythonCodePrompt):
    """Prompt to explain code generation by the LLM"""

    _path_to_template = "assets/prompt_templates/direct_sql_connector.tmpl"

    def _prepare_tables_data(self, tables, config):
        tables_join = []
        for index, table in enumerate(tables):
            table_serialized = DataframeSerializer().serialize(
                table,
                {
                    "index": index,
                    "type": "sql" if config and config.direct_sql else "pandas",
                },
                (
                    config.dataframe_serializer
                    if config
                    else DataframeSerializerType.SQL
                ),
            )
            tables_join.append(table_serialized)
        return "\n\n".join(tables_join)

    def setup(self, tables, config=None, **kwargs) -> None:
        self.set_var("tables", self._prepare_tables_data(tables, config))

        super(DirectSQLPrompt, self).setup(**kwargs)

        self.set_var("current_code", kwargs.pop("current_code", CurrentCodePrompt()))
        self.set_var(
            "code_description",
            kwargs.pop("code_description", "Update this initial code:"),
        )
        self.set_var("last_message", kwargs.pop("last_message", ""))
        self.set_var("prev_conversation", kwargs.pop("prev_conversation", ""))

    def on_prompt_generation(self) -> None:
        default_import = f"import {pd.__name__} as pd"

        self.set_var("default_import", default_import)
        self.set_var("reasoning", SimpleReasoningPrompt())
